from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.id_dict import BatchSpec  # noqa: TC001 # FIXME CoP
from great_expectations.execution_engine.partition_and_sample.data_sampler import (
    DataSampler,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect

if TYPE_CHECKING:
    from great_expectations.compatibility import sqlalchemy
    from great_expectations.execution_engine import SqlAlchemyExecutionEngine


class SqlAlchemyDataSampler(DataSampler):
    """Sampling methods for data stores with SQL interfaces."""

    def sample_using_limit(
        self,
        execution_engine: SqlAlchemyExecutionEngine,
        batch_spec: BatchSpec,
        where_clause: Optional[sqlalchemy.Selectable] = None,
    ) -> Union[str, sqlalchemy.BinaryExpression, sqlalchemy.BooleanClauseList]:
        """Sample using a limit with configuration provided via the batch_spec.

        Note: where_clause needs to be included at this stage since SqlAlchemy's semantics
        for LIMIT are different than normal WHERE clauses.

        Also this requires an engine to find the dialect since certain databases require
        different handling.

        Args:
            execution_engine: Engine used to connect to the database.
            batch_spec: Batch specification describing the batch of interest.
            where_clause: Optional clause used in WHERE clause. Typically generated by a partitioner.

        Returns:
            A query as a string or sqlalchemy object.
        """  # noqa: E501 # FIXME CoP

        # Partition clause should be permissive of all values if not supplied.
        if where_clause is None:
            if execution_engine.dialect_name == GXSqlDialect.SQLITE:
                where_clause = sa.text("1 = 1")  # type: ignore[assignment] # FIXME CoP
            else:
                where_clause = sa.true()  # type: ignore[assignment] # FIXME CoP

        table_name: str = batch_spec["table_name"]

        # SQLalchemy's semantics for LIMIT are different than normal WHERE clauses,
        # so the business logic for building the query needs to be different.
        dialect_name: str = execution_engine.dialect_name
        if dialect_name == GXSqlDialect.ORACLE:
            # TODO: AJB 20220429 WARNING THIS oracle dialect METHOD IS NOT COVERED BY TESTS
            # limit doesn't compile properly for oracle so we will append rownum to query string later  # noqa: E501 # FIXME CoP
            raw_query: sqlalchemy.Selectable = (
                sa.select("*")
                .select_from(sa.table(table_name, schema=batch_spec.get("schema_name", None)))
                .where(where_clause)  # type: ignore[arg-type] # FIXME CoP
            )
            query: str = str(
                raw_query.compile(
                    dialect=execution_engine.dialect,
                    compile_kwargs={"literal_binds": True},
                )
            )
            query += "\nAND ROWNUM <= %d" % batch_spec["sampling_kwargs"]["n"]  # noqa: UP031 # FIXME CoP
            return query
        elif dialect_name == GXSqlDialect.MSSQL:
            # Note that this code path exists because the limit parameter is not getting rendered
            # successfully in the resulting mssql query.
            selectable_query: sqlalchemy.Selectable = (
                sa.select("*")
                .select_from(sa.table(table_name, schema=batch_spec.get("schema_name", None)))
                .where(where_clause)  # type: ignore[arg-type] # FIXME CoP
                .limit(batch_spec["sampling_kwargs"]["n"])
            )
            string_of_query: str = str(
                selectable_query.compile(
                    dialect=execution_engine.dialect,
                    compile_kwargs={"literal_binds": True},
                )
            )
            n: Union[str, int] = batch_spec["sampling_kwargs"]["n"]
            self._validate_mssql_limit_param(n)
            # This string replacement is here because the limit parameter is not substituted during query.compile()  # noqa: E501 # FIXME CoP
            string_of_query = string_of_query.replace("?", str(n))
            return string_of_query
        else:
            return (
                sa.select("*")  # type: ignore[return-value] # FIXME CoP
                .select_from(sa.table(table_name, schema=batch_spec.get("schema_name", None)))
                .where(where_clause)  # type: ignore[arg-type] # FIXME CoP
                .limit(batch_spec["sampling_kwargs"]["n"])
            )

    @staticmethod
    def _validate_mssql_limit_param(n: Union[str, int]) -> None:
        """Validate that the mssql limit param is passed as an int or a string representation of an int.

        Args:
            n: mssql limit parameter.

        Returns:
            None
        """  # noqa: E501 # FIXME CoP
        if not isinstance(n, (str, int)):
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                "Please specify your sampling kwargs 'n' parameter as a string or int."
            )
        if isinstance(n, str) and not n.isdigit():
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                "If specifying your sampling kwargs 'n' parameter as a string please ensure it is "
                "parseable as an integer."
            )

    @staticmethod
    def sample_using_random(
        execution_engine: SqlAlchemyExecutionEngine,
        batch_spec: BatchSpec,
        where_clause: Optional[sqlalchemy.Selectable] = None,
    ) -> sqlalchemy.Selectable:
        """Sample using random data with configuration provided via the batch_spec.

        Note: where_clause needs to be included at this stage since we use the where clause
        to determine the total number of rows to use in determining the rows returned in the
        sample fraction.

        Args:
            execution_engine: Engine used to connect to the database.
            batch_spec: Batch specification describing the batch of interest.
            where_clause: Optional clause used in WHERE clause. Typically generated by a partitioner.

        Returns:
            Sqlalchemy selectable.
        """  # noqa: E501 # FIXME CoP
        try:
            table_name: str = batch_spec["table_name"]
        except KeyError as e:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "A table name must be specified when using sample_using_random. "
                "Please update your configuration"
            ) from e
        try:
            p: float = batch_spec["sampling_kwargs"]["p"] or 1.0
        except (KeyError, TypeError) as e:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "To use sample_using_random you must specify the parameter 'p' in "
                "the 'sampling_kwargs' configuration."
            ) from e

        num_rows: int = execution_engine.execute_query(  # type: ignore[assignment] # FIXME CoP
            sa.select(sa.func.count())
            .select_from(sa.table(table_name, schema=batch_spec.get("schema_name", None)))
            .where(where_clause)  # type: ignore[arg-type] # FIXME CoP
        ).scalar()
        sample_size: int = round(p * num_rows)
        return (
            sa.select("*")
            .select_from(sa.table(table_name, schema=batch_spec.get("schema_name", None)))
            .where(where_clause)  # type: ignore[arg-type] # FIXME CoP
            .order_by(sa.func.random())
            .limit(sample_size)
        )

    def sample_using_mod(
        self,
        batch_spec: BatchSpec,
    ) -> sqlalchemy.Selectable:
        """Take the mod of named column, and only keep rows that match the given value.

        Args:
            batch_spec: should contain keys `column_name`, `mod` and `value`

        Returns:
            Sampled selectable

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("mod", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("value", batch_spec)
        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        mod: int = self.get_sampling_kwargs_value_or_default(batch_spec, "mod")
        value: int = self.get_sampling_kwargs_value_or_default(batch_spec, "value")

        return sa.column(column_name) % mod == value  # type: ignore[return-value] # FIXME CoP

    def sample_using_a_list(
        self,
        batch_spec: BatchSpec,
    ) -> sqlalchemy.Selectable:
        """Match the values in the named column against value_list, and only keep the matches.

        Args:
            batch_spec: should contain keys `column_name` and `value_list`

        Returns:
            Sampled selectable

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("value_list", batch_spec)
        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        value_list: list = self.get_sampling_kwargs_value_or_default(batch_spec, "value_list")
        return sa.column(column_name).in_(value_list)  # type: ignore[return-value] # FIXME CoP

    def sample_using_md5(
        self,
        batch_spec: BatchSpec,
    ) -> sqlalchemy.Selectable:
        """Hash the values in the named column using md5, and only keep rows that match the given hash_value.

        Args:
            batch_spec: should contain keys `column_name` and optionally `hash_digits`
                (default is 1 if not provided), `hash_value` (default is "f" if not provided)

        Returns:
            Sampled selectable

        Raises:
            SamplerError
        """  # noqa: E501 # FIXME CoP
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        hash_digits: int = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec, sampling_kwargs_key="hash_digits", default_value=1
        )
        hash_value: str = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec, sampling_kwargs_key="hash_value", default_value="f"
        )

        return (
            sa.func.right(sa.func.md5(sa.cast(sa.column(column_name), sa.Text)), hash_digits)  # type: ignore[return-value] # FIXME CoP
            == hash_value
        )



from __future__ import annotations

import copy
import pathlib
import random
import string
import urllib.parse
from typing import TYPE_CHECKING, Any
from unittest.mock import ANY as MOCK_ANY
from unittest.mock import patch as mock_patch

import pytest

from great_expectations import get_context
from great_expectations.core.config_provider import _ConfigurationProvider
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.data_context_variables import (
    CloudDataContextVariables,
    DataContextVariables,
    DataContextVariableSchema,
    EphemeralDataContextVariables,
    FileDataContextVariables,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,
    GXCloudConfig,
    ProgressBarsConfig,
)
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
)

if TYPE_CHECKING:
    from unittest.mock import MagicMock  # noqa: TID251 # FIXME CoP

    from pytest_mock import MockerFixture

yaml = YAMLHandler()


@pytest.fixture
def data_context_config_dict() -> dict:
    config: dict = {
        "config_version": 3.0,
        "plugins_directory": "plugins/",
        "validation_results_store_name": "validation_results_store",
        "expectations_store_name": "expectations_store",
        "checkpoint_store_name": "checkpoint_store",
        "config_variables_file_path": "uncommitted/config_variables.yml",
        "stores": {
            "expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "class_name": "TupleFilesystemStoreBackend",
                    "base_directory": "expectations/",
                },
            },
        },
        "data_docs_sites": {},
        "analytics_enabled": True,
        "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        "progress_bars": None,
    }
    return config


@pytest.fixture
def data_context_config(data_context_config_dict: dict) -> DataContextConfig:
    config: DataContextConfig = DataContextConfig(**data_context_config_dict)
    return config


class StubConfigurationProvider(_ConfigurationProvider):
    def __init__(self, config_values=None) -> None:
        self._config_values = config_values or {}
        super().__init__()

    def get_values(self):
        return self._config_values


@pytest.fixture
def ephemeral_data_context_variables(
    data_context_config: DataContextConfig,
) -> EphemeralDataContextVariables:
    return EphemeralDataContextVariables(
        config=data_context_config, config_provider=StubConfigurationProvider()
    )


@pytest.fixture
def file_data_context_variables(
    data_context_config: DataContextConfig, empty_data_context: FileDataContext
) -> FileDataContextVariables:
    return FileDataContextVariables(
        data_context=empty_data_context,
        config=data_context_config,
        config_provider=StubConfigurationProvider(),
    )


@pytest.fixture
def cloud_data_context_variables(
    data_context_config: DataContextConfig,
    ge_cloud_base_url: str,
    ge_cloud_organization_id: str,
    ge_cloud_access_token: str,
) -> CloudDataContextVariables:
    return CloudDataContextVariables(
        ge_cloud_base_url=ge_cloud_base_url,
        ge_cloud_organization_id=ge_cloud_organization_id,
        ge_cloud_access_token=ge_cloud_access_token,
        config=data_context_config,
        config_provider=StubConfigurationProvider(),
    )


@pytest.fixture
def file_data_context(
    tmp_path: pathlib.Path, data_context_config: DataContextConfig
) -> FileDataContext:
    project_path = tmp_path / "file_data_context"
    project_path.mkdir()
    context_root_dir = project_path / FileDataContext.GX_DIR
    context = FileDataContext(project_config=data_context_config, context_root_dir=context_root_dir)
    project_manager.set_project(context)
    return context


@pytest.fixture
def cloud_data_context(
    tmp_path: pathlib.Path,
    data_context_config: DataContextConfig,
    ge_cloud_config_e2e: GXCloudConfig,
) -> CloudDataContext:
    project_path = tmp_path / "cloud_data_context"
    project_path.mkdir()
    context_root_dir = project_path / FileDataContext.GX_DIR

    cloud_data_context = CloudDataContext(
        project_config=data_context_config,
        cloud_base_url=ge_cloud_config_e2e.base_url,
        cloud_access_token=ge_cloud_config_e2e.access_token,
        cloud_organization_id=ge_cloud_config_e2e.organization_id,
        context_root_dir=context_root_dir,
    )
    project_manager.set_project(cloud_data_context)
    return cloud_data_context


def stores() -> dict:
    return {
        "checkpoint_store": {
            "class_name": "CheckpointStore",
            "store_backend": {
                "class_name": "TupleFilesystemStoreBackend",
                "base_directory": "checkpoints/",
            },
        },
    }


@pytest.fixture
def data_docs_sites() -> dict:
    return {
        "local_site": {
            "class_name": "SiteBuilder",
            "show_how_to_buttons": True,
            "store_backend": {
                "class_name": "TupleFilesystemStoreBackend",
                "base_directory": "uncommitted/data_docs/local_site/",
            },
        }
    }


@pytest.fixture
def progress_bars() -> ProgressBarsConfig:
    return ProgressBarsConfig(
        globally=True,
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "target_attr",
    [
        pytest.param(
            DataContextVariableSchema.CONFIG_VERSION,
            id="config_version getter",
        ),
        pytest.param(
            DataContextVariableSchema.CONFIG_VARIABLES_FILE_PATH,
            id="config_variables_file_path getter",
        ),
        pytest.param(
            DataContextVariableSchema.PLUGINS_DIRECTORY,
            id="plugins_directory getter",
        ),
        pytest.param(
            DataContextVariableSchema.EXPECTATIONS_STORE_NAME,
            id="expectations_store getter",
        ),
        pytest.param(
            DataContextVariableSchema.VALIDATIONS_STORE_NAME,
            id="validation_results_store getter",
        ),
        pytest.param(
            DataContextVariableSchema.CHECKPOINT_STORE_NAME,
            id="checkpoint_store getter",
        ),
        pytest.param(DataContextVariableSchema.STORES, id="stores getter"),
        pytest.param(
            DataContextVariableSchema.DATA_DOCS_SITES,
            id="data_docs_sites getter",
        ),
        pytest.param(
            DataContextVariableSchema.PROGRESS_BARS,
            id="progress_bars getter",
        ),
    ],
)
@pytest.mark.slow  # 1.20s
def test_data_context_variables_get(
    ephemeral_data_context_variables: EphemeralDataContextVariables,
    file_data_context_variables: FileDataContextVariables,
    cloud_data_context_variables: CloudDataContextVariables,
    data_context_config: dict,
    target_attr: DataContextVariableSchema,
) -> None:
    def _test_variables_get(type_: DataContextVariables) -> None:
        res: Any = getattr(type_, target_attr.value)

        expected_value: Any = data_context_config[target_attr.value]
        assert res == expected_value

    # EphemeralDataContextVariables
    _test_variables_get(ephemeral_data_context_variables)

    # FileDataContextVariables
    _test_variables_get(file_data_context_variables)

    # CloudDataContextVariables
    _test_variables_get(cloud_data_context_variables)


@pytest.mark.unit
def test_data_context_variables_get_with_substitutions(
    data_context_config_dict: dict,
) -> None:
    env_var_name: str = "MY_CONFIG_VERSION"
    value_associated_with_env_var: float = 7.0

    data_context_config_dict[DataContextVariableSchema.CONFIG_VERSION] = f"${env_var_name}"
    config: DataContextConfig = DataContextConfig(**data_context_config_dict)
    config_values: dict = {
        env_var_name: value_associated_with_env_var,
    }
    variables: DataContextVariables = EphemeralDataContextVariables(
        config=config,
        config_provider=StubConfigurationProvider(config_values=config_values),
    )
    assert variables.config_version == value_associated_with_env_var


@pytest.mark.unit
@pytest.mark.parametrize(
    "input_value,target_attr",
    [
        pytest.param(
            5.0,
            DataContextVariableSchema.CONFIG_VERSION,
            id="config_version setter",
        ),
        pytest.param(
            "uncommitted/my_config_file.yml",
            DataContextVariableSchema.CONFIG_VARIABLES_FILE_PATH,
            id="config_variables_file_path setter",
        ),
        pytest.param(
            "other_plugins/",
            DataContextVariableSchema.PLUGINS_DIRECTORY,
            id="plugins_directory setter",
        ),
        pytest.param(
            "my_expectations_store",
            DataContextVariableSchema.EXPECTATIONS_STORE_NAME,
            id="expectations_store setter",
        ),
        pytest.param(
            "my_validation_results_store",
            DataContextVariableSchema.VALIDATIONS_STORE_NAME,
            id="validation_results_store setter",
        ),
        pytest.param(
            "my_checkpoint_store",
            DataContextVariableSchema.CHECKPOINT_STORE_NAME,
            id="checkpoint_store setter",
        ),
        pytest.param(stores, DataContextVariableSchema.STORES, id="stores setter"),
        pytest.param(
            data_docs_sites,
            DataContextVariableSchema.DATA_DOCS_SITES,
            id="data_docs_sites setter",
        ),
        pytest.param(
            progress_bars,
            DataContextVariableSchema.PROGRESS_BARS,
            id="progress_bars setter",
        ),
    ],
)
@pytest.mark.slow  # 1.20s
def test_data_context_variables_set(
    ephemeral_data_context_variables: EphemeralDataContextVariables,
    file_data_context_variables: FileDataContextVariables,
    cloud_data_context_variables: CloudDataContextVariables,
    input_value: Any,
    target_attr: DataContextVariableSchema,
) -> None:
    def _test_variables_set(type_: DataContextVariables) -> None:
        setattr(type_, target_attr.value, input_value)
        res: Any = type_.config[target_attr.value]

        assert res == input_value

    # EphemeralDataContextVariables
    _test_variables_set(ephemeral_data_context_variables)

    # FileDataContextVariables
    _test_variables_set(file_data_context_variables)

    # CloudDataContextVariables
    _test_variables_set(cloud_data_context_variables)


@pytest.mark.unit
def test_data_context_variables_save(
    mocker: MockerFixture,
    data_context_config_dict: dict,
    ephemeral_data_context_variables: EphemeralDataContextVariables,
    file_data_context_variables: FileDataContextVariables,
    cloud_data_context_variables: CloudDataContextVariables,
    # The below GX Cloud variables were used to instantiate the above CloudDataContextVariables
    v1_cloud_base_url: str,
    ge_cloud_organization_id: str,
    ge_cloud_access_token: str,
) -> None:
    # EphemeralDataContextVariables
    ephemeral_data_context_variables.save()
    key: ConfigurationIdentifier = ephemeral_data_context_variables.get_key()
    persisted_value: DataContextConfig = ephemeral_data_context_variables.store.get(key=key)
    assert persisted_value.to_json_dict() == ephemeral_data_context_variables.config.to_json_dict()

    # FileDataContextVariables
    mock_save = mocker.patch(
        "great_expectations.data_context.store.InlineStoreBackend._save_changes",
        autospec=True,
    )
    file_data_context_variables.save()

    assert mock_save.call_count == 1

    # CloudDataContextVariables
    mock_put = mocker.patch("requests.Session.put", autospec=True)
    type(mock_put.return_value).status_code = mocker.PropertyMock(return_value=200)

    cloud_data_context_variables.save()

    expected_config_dict = {
        "analytics_enabled": True,
        "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        "config_variables_file_path": "uncommitted/config_variables.yml",
        "config_version": 3.0,
        "data_docs_sites": {},
        "plugins_directory": "plugins/",
        "stores": {
            "expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "class_name": "TupleFilesystemStoreBackend",
                    "base_directory": "expectations/",
                },
            },
            "checkpoint_store": {"class_name": "CheckpointStore"},
            "validation_results_store": {"class_name": "ValidationResultsStore"},
            "validation_definition_store": {"class_name": "ValidationDefinitionStore"},
        },
    }

    assert mock_put.call_count == 1
    url = urllib.parse.urljoin(
        v1_cloud_base_url, f"organizations/{ge_cloud_organization_id}/data-context-variables"
    )
    mock_put.assert_called_with(
        MOCK_ANY,  # requests.Session object
        url,
        json={
            "data": expected_config_dict,
        },
    )


@pytest.mark.unit
def test_data_context_variables_repr_and_str_only_reveal_config(
    data_context_config: DataContextConfig,
) -> None:
    config = data_context_config

    variables = EphemeralDataContextVariables(
        config=data_context_config,
        config_provider=StubConfigurationProvider(),
    )

    variables_str = str(variables)
    variables_repr = repr(variables)

    assert variables_str == str(config)
    assert variables_repr == repr(config)


@pytest.mark.big
def test_file_data_context_variables_e2e(
    monkeypatch,
    file_data_context: FileDataContext,
    progress_bars: ProgressBarsConfig,
) -> None:
    """
    What does this test do and why?

    Tests the E2E workflow with a FileDataContextVariables instance.
      1. User updates certain values and sets them as attributes.
      2. User persists changes utilizing the save call defined by the Variables API.
      3. Upon reading the result config from disk, we can confirm that changes were appropriately persisted.

    It is also important to note that in the case of $VARS syntax, we NEVER want to persist the underlying
    value in order to preserve sensitive information.
    """  # noqa: E501 # FIXME CoP
    # Prepare updated progress_bars to set and serialize to disk
    updated_progress_bars: ProgressBarsConfig = copy.deepcopy(progress_bars)
    updated_progress_bars.globally = False

    # Prepare updated plugins directory to set and serialize to disk (ensuring we hide the true value behind $VARS syntax)  # noqa: E501 # FIXME CoP
    env_var_name: str = "MY_PLUGINS_DIRECTORY"
    value_associated_with_env_var: str = "foo/bar/baz"
    monkeypatch.setenv(env_var_name, value_associated_with_env_var)

    # Set attributes defined above
    file_data_context.variables.progress_bars = updated_progress_bars
    file_data_context.variables.plugins_directory = f"${env_var_name}"
    file_data_context.variables.save()

    # Review great_expectations.yml where values were written and confirm changes
    config_filepath = pathlib.Path(file_data_context.root_directory).joinpath(
        file_data_context.GX_YML
    )

    with open(config_filepath) as f:
        contents: dict = yaml.load(f)
        config_saved_to_disk: DataContextConfig = DataContextConfig(**contents)

    assert config_saved_to_disk.progress_bars == updated_progress_bars.to_dict()
    assert file_data_context.variables.plugins_directory == value_associated_with_env_var
    assert config_saved_to_disk.plugins_directory == f"${env_var_name}"


@pytest.mark.cloud
@pytest.mark.xfail(
    strict=False,
    reason="GX Cloud E2E tests are failing due to new top-level `analytics` and `data_context_id` variables not yet being recognized by the server",  # noqa: E501 # FIXME CoP
)
def test_cloud_data_context_variables_successfully_hits_cloud_endpoint(
    cloud_data_context: CloudDataContext,
    data_context_config: DataContextConfig,
) -> None:
    """
    What does this test do and why?

    Ensures that the endpoint responsible for the DataContextVariables resource is accessible
    through the Variables API.
    """
    cloud_data_context.variables.config = data_context_config
    success = cloud_data_context.variables.save()

    assert success is True


@pytest.mark.cloud
@mock_patch(
    "great_expectations.data_context.data_context.serializable_data_context.SerializableDataContext._save_project_config"
)
@pytest.mark.xfail(
    strict=False,
    reason="GX Cloud E2E tests are failing due to env vars not being consistently recognized by Docker; x-failing for purposes of 0.15.22 release",  # noqa: E501 # FIXME CoP
)
def test_cloud_enabled_data_context_variables_e2e(
    mock_save_project_config: MagicMock,
    data_docs_sites: dict,
    monkeypatch,
) -> None:
    """
    What does this test do and why?

    Tests the E2E workflow with a Cloud-enabled DataContext; as the CloudDataContext does not yet have 1-to-1
    feature parity with the DataContext (as v0.15.15), this is the primary mechanism by which Great
    Expectations Cloud interacts with variables.
      1. User updates certain values and sets them as attributes.
      2. User persists changes utilizing the save call defined by the Variables API.
      3. Upon reading the result config from a GET request, we can confirm that changes were appropriately persisted.

    It is also important to note that in the case of $VARS syntax, we NEVER want to persist the underlying
    value in order to preserve sensitive information.
    """  # noqa: E501 # FIXME CoP
    # Prepare updated plugins directory to set and save to the Cloud backend.
    # As values are persisted in the Cloud DB, we want to randomize our values each time for consistent test results  # noqa: E501 # FIXME CoP
    updated_plugins_dir = f"plugins_dir_{''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))}"  # noqa: E501 # FIXME CoP

    updated_data_docs_sites = data_docs_sites
    new_site_name = f"docs_site_{''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))}"  # noqa: E501 # FIXME CoP
    updated_data_docs_sites[new_site_name] = {}

    context = get_context(cloud_mode=True)

    assert context.variables.plugins_directory != updated_plugins_dir
    assert context.variables.data_docs_sites != updated_data_docs_sites

    context.variables.plugins_directory = updated_plugins_dir
    context.variables.data_docs_sites = updated_data_docs_sites

    assert context.variables.plugins_directory == updated_plugins_dir
    assert context.variables.data_docs_sites == updated_data_docs_sites

    context.variables.save()

    context = get_context(cloud_mode=True)

    assert context.variables.plugins_directory == updated_plugins_dir
    assert context.variables.data_docs_sites == updated_data_docs_sites



import great_expectations as gx

context = gx.get_context()

datasource_name = "my_s3_datasource"
bucket_name = "great-expectations-docs-test"

datasource = context.data_sources.add_pandas_s3(
    name=datasource_name, bucket=bucket_name, boto3_options={}
)

asset_name = "my_taxi_data_asset"
s3_prefix = "data/taxi_yellow_tripdata_samples/"
data_asset = datasource.add_csv_asset(name=asset_name, s3_prefix=s3_prefix)

batch_definition = data_asset.add_batch_definition_path(
    "path",
    path="yellow_tripdata_sample_2019-03.csv",
)

batch_request = batch_definition.build_batch_request()
batch_list = data_asset.get_batch(batch_request)

assert batch_list.metadata == {
    "path": "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-03.csv"
}



from __future__ import annotations

from typing import Literal, Optional

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.directory_asset import (
    DirectoryDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.data_asset.path.spark.spark_generic import (
    _SparkGenericFilePathAssetMixin,
)


class TextAssetBase(_SparkGenericFilePathAssetMixin):
    # The options below are available as of spark v3.4.0
    # See https://spark.apache.org/docs/latest/sql-data-sources-text.html for more info.
    wholetext: bool = Field(False)
    line_sep: Optional[str] = Field(None, alias="lineSep")

    class Config:
        extra = "forbid"

        allow_population_by_field_name = True

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "text"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-text.html for more info.
        """
        return super()._get_reader_options_include().union({"wholetext", "line_sep"})


class TextAsset(FileDataAsset, TextAssetBase):
    type: Literal["text"] = "text"


class DirectoryTextAsset(DirectoryDataAsset, TextAssetBase):
    type: Literal["directory_text"] = "directory_text"

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "text"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-text.html for more info.
        """
        return (
            super()._get_reader_options_include()
            | super(DirectoryDataAsset, self)._get_reader_options_include()
        )



# PYTHON 2 - py2 - update to ABC direct use rather than __metaclass__ once we drop py2 support
from __future__ import annotations

import functools
import logging
import os
import pathlib
import random
import re
import shutil
from abc import ABCMeta
from pathlib import Path
from typing import Any, List, Tuple

from great_expectations.compatibility import aws
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.exceptions import InvalidKeyError, StoreBackendError
from great_expectations.util import filter_properties_dict

logger = logging.getLogger(__name__)


class TupleStoreBackend(StoreBackend, metaclass=ABCMeta):
    r"""
    If filepath_template is provided, the key to this StoreBackend abstract class must be a tuple with
    fixed length equal to the number of unique components matching the regex r"{\d+}"

    For example, in the following template path: expectations/{0}/{1}/{2}/prefix-{2}.json, keys must have
    three components.
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        filepath_template=None,
        filepath_prefix=None,
        filepath_suffix=None,
        forbidden_substrings=None,
        platform_specific_separator=True,
        fixed_length_key=False,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        base_public_path=None,
        store_name=None,
    ) -> None:
        super().__init__(
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            store_name=store_name,
        )
        if forbidden_substrings is None:
            forbidden_substrings = ["/", "\\"]
        self.forbidden_substrings = forbidden_substrings
        self.platform_specific_separator = platform_specific_separator

        if filepath_template is not None and filepath_suffix is not None:
            raise ValueError("filepath_suffix may only be used when filepath_template is None")  # noqa: TRY003 # FIXME CoP

        self.filepath_template = filepath_template
        if filepath_prefix and len(filepath_prefix) > 0:
            # Validate that the filepath prefix does not end with a forbidden substring
            if filepath_prefix[-1] in self.forbidden_substrings:
                raise StoreBackendError(
                    "Unable to initialize TupleStoreBackend: filepath_prefix may not end with a "
                    "forbidden substring. Current forbidden substrings are "
                    + str(forbidden_substrings)
                )
        self.filepath_prefix = filepath_prefix
        self.filepath_suffix = filepath_suffix
        self.base_public_path = base_public_path

        if filepath_template is not None:
            # key length is the number of unique values to be substituted in the filepath_template
            self.key_length = len(set(re.findall(r"{\d+}", filepath_template)))

            self.verify_that_key_to_filepath_operation_is_reversible()
            self._fixed_length_key = True

    @staticmethod
    def _is_missing_prefix_or_suffix(filepath_prefix: str, filepath_suffix: str, key: str) -> bool:
        missing_prefix = bool(filepath_prefix and not key.startswith(filepath_prefix))
        missing_suffix = bool(filepath_suffix and not key.endswith(filepath_suffix))
        return missing_prefix or missing_suffix

    @override
    def _validate_key(self, key) -> None:
        super()._validate_key(key)

        for key_element in key:
            for substring in self.forbidden_substrings:
                if substring in key_element:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        f"Keys in {self.__class__.__name__} must not contain substrings in {self.forbidden_substrings} : {key}"  # noqa: E501 # FIXME CoP
                    )

    @override
    def _validate_value(self, value) -> None:
        if not isinstance(value, str) and not isinstance(value, bytes):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"Values in {self.__class__.__name__} must be instances of {str} or {bytes}, not {type(value)}"  # noqa: E501 # FIXME CoP
            )

    def _convert_key_to_filepath(self, key):
        # NOTE: This method uses a hard-coded forward slash as a separator,
        # and then replaces that with a platform-specific separator if requested (the default)
        self._validate_key(key)
        # Handle store_backend_id separately
        if key == self.STORE_BACKEND_ID_KEY:
            filepath = f"{self.filepath_prefix or ''}{'/' if self.filepath_prefix else ''}{key[0]}"
            return filepath if not self.platform_specific_separator else os.path.normpath(filepath)
        if self.filepath_template:
            converted_string = self.filepath_template.format(*list(key))
        else:
            converted_string = "/".join(key)

        if self.filepath_prefix:
            converted_string = f"{self.filepath_prefix}/{converted_string}"
        if self.filepath_suffix:
            converted_string += self.filepath_suffix
        if self.platform_specific_separator:
            converted_string = os.path.normpath(converted_string)

        return converted_string

    def _convert_filepath_to_key(self, filepath):  # noqa: C901, PLR0912 # FIXME CoP
        if filepath == self.STORE_BACKEND_ID_KEY[0]:
            return self.STORE_BACKEND_ID_KEY
        if self.platform_specific_separator:
            filepath = os.path.normpath(filepath)

        if self.filepath_prefix:
            if (
                not filepath.startswith(self.filepath_prefix)
                and len(filepath) >= len(self.filepath_prefix) + 1
            ):
                # If filepath_prefix is set, we expect that it is the first component of a valid filepath.  # noqa: E501 # FIXME CoP
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "filepath must start with the filepath_prefix when one is set by the store_backend"  # noqa: E501 # FIXME CoP
                )
            else:
                # Remove the prefix before processing
                # Also remove the separator that was added, which may have been platform-dependent
                filepath = filepath[len(self.filepath_prefix) + 1 :]

        if self.filepath_suffix:
            if not filepath.endswith(self.filepath_suffix):
                # If filepath_suffix is set, we expect that it is the last component of a valid filepath.  # noqa: E501 # FIXME CoP
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "filepath must end with the filepath_suffix when one is set by the store_backend"  # noqa: E501 # FIXME CoP
                )
            else:
                # Remove the suffix before processing
                filepath = filepath[: -len(self.filepath_suffix)]

        if self.filepath_template:
            # filepath_template is always specified with forward slashes, but it is then
            # used to (1) dynamically construct and evaluate a regex, and (2) split the provided (observed) filepath  # noqa: E501 # FIXME CoP
            if self.platform_specific_separator:
                filepath_template = os.path.join(  # noqa: PTH118 # FIXME CoP
                    *self.filepath_template.split("/")
                )
                filepath_template = filepath_template.replace("\\", "\\\\")
            else:
                filepath_template = self.filepath_template

            # Convert the template to a regex
            indexed_string_substitutions = re.findall(r"{\d+}", filepath_template)
            tuple_index_list = [
                f"(?P<tuple_index_{i}>.*)" for i in range(len(indexed_string_substitutions))
            ]
            intermediate_filepath_regex = re.sub(
                r"{\d+}",
                lambda m,
                r=iter(  # noqa: B008 # function-call-in-default-argument
                    tuple_index_list
                ): next(r),
                filepath_template,
            )
            filepath_regex = intermediate_filepath_regex.format(*tuple_index_list)

            # Apply the regex to the filepath
            matches = re.compile(filepath_regex).match(filepath)
            if matches is None:
                return None

            # Map key elements into the appropriate parts of the tuple
            new_key = [None] * self.key_length
            for i in range(len(tuple_index_list)):
                tuple_index = int(re.search(r"\d+", indexed_string_substitutions[i]).group(0))
                key_element = matches.group(f"tuple_index_{i!s}")
                new_key[tuple_index] = key_element

            new_key = tuple(new_key)
        else:
            new_key = pathlib.Path(filepath).parts
        return new_key

    def verify_that_key_to_filepath_operation_is_reversible(self):
        def get_random_hex(size=4):
            return "".join([random.choice(list("ABCDEF0123456789")) for _ in range(size)])

        key = tuple(get_random_hex() for _ in range(self.key_length))
        filepath = self._convert_key_to_filepath(key)
        new_key = self._convert_filepath_to_key(filepath)
        if key != new_key:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"filepath template {self.filepath_template} for class {self.__class__.__name__} is not reversible for a tuple of length {self.key_length}. "  # noqa: E501 # FIXME CoP
                "Have you included all elements in the key tuple?"
            )

    @property
    @override
    def config(self) -> dict:
        return self._config  # type: ignore[attr-defined] # FIXME CoP


class TupleFilesystemStoreBackend(TupleStoreBackend):
    """Uses a local filepath as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.
    The filepath_template is a string template used to convert the key to a filepath.
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        base_directory,
        filepath_template=None,
        filepath_prefix=None,
        filepath_suffix=None,
        forbidden_substrings=None,
        platform_specific_separator=True,
        root_directory=None,
        fixed_length_key=False,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        base_public_path=None,
        store_name=None,
    ) -> None:
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            base_public_path=base_public_path,
            store_name=store_name,
        )
        if os.path.isabs(base_directory):  # noqa: PTH117 # FIXME CoP
            self.full_base_directory = base_directory
        else:  # noqa: PLR5501 # FIXME CoP
            if root_directory is None:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "base_directory must be an absolute path if root_directory is not provided"
                )
            elif not os.path.isabs(root_directory):  # noqa: PTH117 # FIXME CoP
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    f"root_directory must be an absolute path. Got {root_directory} instead."
                )
            else:
                self.full_base_directory = os.path.join(  # noqa: PTH118 # FIXME CoP
                    root_directory, base_directory
                )

        os.makedirs(  # noqa: PTH103 # FIXME CoP
            str(os.path.dirname(self.full_base_directory)),  # noqa: PTH120 # FIXME CoP
            exist_ok=True,
        )
        # Initialize with store_backend_id if not part of an HTMLSiteStore
        if not self._suppress_store_backend_id:
            _ = self.store_backend_id

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "base_directory": base_directory,
            "filepath_template": filepath_template,
            "filepath_prefix": filepath_prefix,
            "filepath_suffix": filepath_suffix,
            "forbidden_substrings": forbidden_substrings,
            "platform_specific_separator": platform_specific_separator,
            "root_directory": root_directory,
            "fixed_length_key": fixed_length_key,
            "suppress_store_backend_id": suppress_store_backend_id,
            "manually_initialize_store_backend_id": manually_initialize_store_backend_id,
            "base_public_path": base_public_path,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def _get(self, key):  # type: ignore[explicit-override] # FIXME
        filepath: str = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.full_base_directory, self._convert_key_to_filepath(key)
        )
        try:
            with open(filepath) as infile:
                contents: str = infile.read().rstrip("\n")
        except FileNotFoundError as e:
            raise InvalidKeyError(  # noqa: TRY003 # FIXME CoP
                f"Unable to retrieve object from TupleFilesystemStoreBackend with the following Key: {filepath!s}"  # noqa: E501 # FIXME CoP
            ) from e

        return contents

    @override
    def _get_all(self) -> list[Any]:
        keys = [key for key in self.list_keys() if key != StoreBackend.STORE_BACKEND_ID_KEY]
        return [self._get(key) for key in keys]

    def _set(self, key, value, **kwargs):  # type: ignore[explicit-override] # FIXME
        if not isinstance(key, tuple):
            key = key.to_tuple()
        filepath = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.full_base_directory, self._convert_key_to_filepath(key)
        )
        path, _filename = os.path.split(filepath)

        os.makedirs(str(path), exist_ok=True)  # noqa: PTH103 # FIXME CoP
        with open(filepath, "wb") as outfile:
            if isinstance(value, str):
                outfile.write(value.encode("utf-8"))
            else:
                outfile.write(value)
        return filepath

    def _move(self, source_key, dest_key, **kwargs):  # type: ignore[explicit-override] # FIXME
        source_path = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.full_base_directory, self._convert_key_to_filepath(source_key)
        )

        dest_path = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.full_base_directory, self._convert_key_to_filepath(dest_key)
        )
        dest_dir, _dest_filename = os.path.split(dest_path)

        if os.path.exists(source_path):  # noqa: PTH110 # FIXME CoP
            os.makedirs(dest_dir, exist_ok=True)  # noqa: PTH103 # FIXME CoP
            shutil.move(source_path, dest_path)
            return dest_key

        return False

    @override
    def list_keys(self, prefix: Tuple = ()) -> List[Tuple]:
        key_list = []
        for root, dirs, files in os.walk(
            os.path.join(self.full_base_directory, *prefix)  # noqa: PTH118 # FIXME CoP
        ):
            for file_ in files:
                full_path, file_name = os.path.split(
                    os.path.join(root, file_)  # noqa: PTH118 # FIXME CoP
                )
                relative_path = os.path.relpath(
                    full_path,
                    self.full_base_directory,
                )
                if relative_path == ".":
                    filepath = file_name
                else:
                    filepath = os.path.join(relative_path, file_name)  # noqa: PTH118 # FIXME CoP

                if self._is_missing_prefix_or_suffix(
                    filepath_prefix=self.filepath_prefix,
                    filepath_suffix=self.filepath_suffix,
                    key=filepath,
                ):
                    continue
                key = self._convert_filepath_to_key(filepath)
                if key and not self.is_ignored_key(key):
                    key_list.append(key)

        return key_list

    def rrmdir(self, mroot, curpath) -> None:
        """
        recursively removes empty dirs between curpath and mroot inclusive
        """
        try:
            while (
                not Path(curpath).iterdir()
                and os.path.exists(curpath)  # noqa: PTH110 # FIXME CoP
                and mroot != curpath
            ):
                f2 = os.path.dirname(curpath)  # noqa: PTH120 # FIXME CoP
                os.rmdir(curpath)  # noqa: PTH106 # FIXME CoP
                curpath = f2
        except (NotADirectoryError, FileNotFoundError):
            pass

    def remove_key(self, key):  # type: ignore[explicit-override] # FIXME
        if not isinstance(key, tuple):
            key = key.to_tuple()

        filepath = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.full_base_directory, self._convert_key_to_filepath(key)
        )

        if os.path.exists(filepath):  # noqa: PTH110 # FIXME CoP
            d_path = os.path.dirname(filepath)  # noqa: PTH120 # FIXME CoP
            os.remove(filepath)  # noqa: PTH107 # FIXME CoP
            self.rrmdir(self.full_base_directory, d_path)
            return True
        return False

    @override
    def get_url_for_key(self, key, protocol=None) -> str:
        path = self._convert_key_to_filepath(key)
        escaped_path = self._url_path_escape_special_characters(path=path)
        full_path = os.path.join(self.full_base_directory, escaped_path)  # noqa: PTH118 # FIXME CoP

        if protocol is None:
            protocol = "file:"
        url = f"{protocol}//{full_path}"
        return url

    def get_public_url_for_key(self, key, protocol=None):
        if not self.base_public_path:
            raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                """Error: No base_public_path was configured!
                    - A public URL was requested base_public_path was not configured for the TupleFilesystemStoreBackend
                """  # noqa: E501 # FIXME CoP
            )
        path = self._convert_key_to_filepath(key)
        public_url = self.base_public_path + path
        return public_url

    def _has_key(self, key):  # type: ignore[explicit-override] # FIXME
        return os.path.isfile(  # noqa: PTH113 # FIXME CoP
            os.path.join(  # noqa: PTH118 # FIXME CoP
                self.full_base_directory, self._convert_key_to_filepath(key)
            )
        )

    @property
    @override
    def config(self) -> dict:
        return self._config


class TupleS3StoreBackend(TupleStoreBackend):
    """
    Uses an S3 bucket as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.
    The filepath_template is a string template used to convert the key to a filepath.
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        bucket,
        prefix="",
        boto3_options=None,
        s3_put_options=None,
        filepath_template=None,
        filepath_prefix=None,
        filepath_suffix=None,
        forbidden_substrings=None,
        platform_specific_separator=False,
        fixed_length_key=False,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        base_public_path=None,
        endpoint_url=None,
        store_name=None,
    ) -> None:
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            base_public_path=base_public_path,
            store_name=store_name,
        )
        self.bucket = bucket
        if prefix:
            if self.platform_specific_separator:
                prefix = prefix.strip(os.sep)

            # we *always* strip "/" from the prefix based on the norms of s3
            # whether the rest of the key is built with platform-specific separators or not
            prefix = prefix.strip("/")
        self.prefix = prefix
        if boto3_options is None:
            boto3_options = {}
        self._boto3_options = boto3_options
        if s3_put_options is None:
            s3_put_options = {}
        self.s3_put_options = s3_put_options
        self.endpoint_url = endpoint_url
        # Initialize with store_backend_id if not part of an HTMLSiteStore
        if not self._suppress_store_backend_id:
            _ = self.store_backend_id

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "bucket": bucket,
            "prefix": prefix,
            "boto3_options": boto3_options,
            "s3_put_options": s3_put_options,
            "filepath_template": filepath_template,
            "filepath_prefix": filepath_prefix,
            "filepath_suffix": filepath_suffix,
            "forbidden_substrings": forbidden_substrings,
            "platform_specific_separator": platform_specific_separator,
            "fixed_length_key": fixed_length_key,
            "suppress_store_backend_id": suppress_store_backend_id,
            "manually_initialize_store_backend_id": manually_initialize_store_backend_id,
            "base_public_path = None": base_public_path,
            "endpoint_url": endpoint_url,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def _build_s3_object_key(self, key):
        if self.platform_specific_separator:
            if self.prefix:
                s3_object_key = os.path.join(  # noqa: PTH118 # FIXME CoP
                    self.prefix, self._convert_key_to_filepath(key)
                )
            else:
                s3_object_key = self._convert_key_to_filepath(key)
        else:  # noqa: PLR5501 # FIXME CoP
            if self.prefix:
                s3_object_key = "/".join((self.prefix, self._convert_key_to_filepath(key)))
            else:
                s3_object_key = self._convert_key_to_filepath(key)
        return s3_object_key

    def _get(self, key):  # type: ignore[explicit-override] # FIXME
        client = self._create_client()
        s3_object_key = self._build_s3_object_key(key)
        return self._get_by_s3_object_key(client, s3_object_key)

    @override
    def _get_all(self) -> list[Any]:
        """Get all objects from the store.
        NOTE: This is non-performant because we download each object separately.
        See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/bucket/objects.html#objects
        for the docs.
        """
        client = self._create_client()
        keys = self.list_keys()
        keys = [k for k in keys if k != StoreBackend.STORE_BACKEND_ID_KEY]
        s3_object_keys = [self._build_s3_object_key(key) for key in keys]
        return [self._get_by_s3_object_key(client, key) for key in s3_object_keys]

    def _get_by_s3_object_key(self, s3_client, s3_object_key) -> Any:
        try:
            s3_response_object = s3_client.get_object(Bucket=self.bucket, Key=s3_object_key)
        except (s3_client.exceptions.NoSuchKey, s3_client.exceptions.NoSuchBucket) as e:
            raise InvalidKeyError(  # noqa: TRY003 # FIXME CoP
                f"Unable to retrieve object from TupleS3StoreBackend with the following Key: {s3_object_key!s}"  # noqa: E501 # FIXME CoP
            ) from e

        # ContentEncoding is an unknown string per
        # https://botocore.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/get_object.html#get-object
        content_encoding: str = s3_response_object.get("ContentEncoding", "utf-8")
        # We found the string could take the form of an array, e.g. `utf-8,aws-chunked`.
        # As of boto3 1.36.0, we aren't aware of any time when the string-list can have more
        # than 1 item except when `aws-chunked` is included. In order to preserve unknown
        # encodings included with `aws-chunked` we will remove the `aws-chunked` string from
        # the list. We do not intend to add support for reading in chunks at this time.
        # Calling botocore.response.StreamingBody.read() without arguments
        # will read the entire stream.
        encodings: list[str] = content_encoding.split(",")
        if "aws-chunked" in encodings:
            encodings.remove("aws-chunked")
        data = s3_response_object["Body"].read()
        # Maybe it's theoretically possible to encode an entire file with multiple encodings.
        # It's an odd choice of format for ContentEncoding to allow multiple encoding strings.
        # As far as we know, this list will never end up with length>1 after the possible
        # encoding `aws-chunked` is removed.
        for encoding in encodings:
            data = data.decode(encoding)
        return data

    def _set(  # type: ignore[explicit-override] # FIXME
        self,
        key,
        value,
        content_encoding="utf-8",
        content_type="application/json",
        **kwargs,
    ):
        s3_object_key = self._build_s3_object_key(key)

        s3 = self._create_resource()

        try:
            result_s3 = s3.Object(self.bucket, s3_object_key)
            if isinstance(value, str):
                result_s3.put(
                    Body=value.encode(content_encoding),
                    ContentEncoding=content_encoding,
                    ContentType=content_type,
                    **self.s3_put_options,
                )
            else:
                result_s3.put(Body=value, ContentType=content_type, **self.s3_put_options)
        except s3.meta.client.exceptions.ClientError as e:
            logger.debug(str(e))
            raise StoreBackendError("Unable to set object in s3.") from e  # noqa: TRY003 # FIXME CoP

        return s3_object_key

    @override
    def _move(self, source_key, dest_key, **kwargs) -> None:
        s3 = self._create_resource()

        source_filepath = self._convert_key_to_filepath(source_key)
        if not source_filepath.startswith(self.prefix):
            source_filepath = os.path.join(self.prefix, source_filepath)  # noqa: PTH118 # FIXME CoP
        dest_filepath = self._convert_key_to_filepath(dest_key)
        if not dest_filepath.startswith(self.prefix):
            dest_filepath = os.path.join(self.prefix, dest_filepath)  # noqa: PTH118 # FIXME CoP

        s3.Bucket(self.bucket).copy({"Bucket": self.bucket, "Key": source_filepath}, dest_filepath)

        s3.Object(self.bucket, source_filepath).delete()

    @override
    def list_keys(self, prefix: Tuple = ()) -> List[Tuple]:  # noqa: C901 #  too complex
        # Note that the prefix arg is only included to maintain consistency with the parent class signature  # noqa: E501 # FIXME CoP
        s3r = self._create_resource()
        bucket = s3r.Bucket(self.bucket)
        key_list = []
        if self.prefix:
            objects_list = bucket.objects.filter(Prefix=self.prefix)
        else:
            objects_list = bucket.objects.all()
        for s3_object_info in objects_list:
            s3_object_key = s3_object_info.key
            if self.platform_specific_separator:
                s3_object_key = os.path.relpath(s3_object_key, self.prefix)
            else:  # noqa: PLR5501 # FIXME CoP
                if self.prefix is None:
                    if s3_object_key.startswith("/"):
                        s3_object_key = s3_object_key[1:]
                else:  # noqa: PLR5501 # FIXME CoP
                    if s3_object_key.startswith(f"{self.prefix}/"):
                        s3_object_key = s3_object_key[len(self.prefix) + 1 :]

            if self._is_missing_prefix_or_suffix(
                filepath_prefix=self.filepath_prefix,
                filepath_suffix=self.filepath_suffix,
                key=s3_object_key,
            ):
                continue
            key = self._convert_filepath_to_key(s3_object_key)
            if key:
                key_list.append(key)
        return key_list

    @override
    def get_url_for_key(self, key, protocol=None) -> str:
        location = None
        if self.boto3_options.get("endpoint_url"):
            location = self.boto3_options.get("endpoint_url")
        else:
            # build s3 endpoint when no endpoint_url is configured

            location = self._create_client().get_bucket_location(Bucket=self.bucket)[
                "LocationConstraint"
            ]

            if location is None:
                location = "https://s3.amazonaws.com"
            else:
                location = f"https://s3-{location}.amazonaws.com"

        s3_key = self._convert_key_to_filepath(key)

        if not self.prefix:
            return f"{location}/{self.bucket}/{s3_key}"
        return f"{location}/{self.bucket}/{self.prefix}/{s3_key}"

    def get_public_url_for_key(self, key, protocol=None):
        if not self.base_public_path:
            raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                """Error: No base_public_path was configured!
                    - A public URL was requested base_public_path was not configured for the
                """
            )
        s3_key = self._convert_key_to_filepath(key)
        # <WILL> What happens if there is a prefix?
        if self.base_public_path[-1] != "/":
            public_url = f"{self.base_public_path}/{s3_key}"
        else:
            public_url = self.base_public_path + s3_key
        return public_url

    def remove_key(self, key):  # type: ignore[explicit-override] # FIXME
        if not isinstance(key, tuple):
            key = key.to_tuple()

        s3 = self._create_resource()
        s3_object_key = self._build_s3_object_key(key)

        # Check if the object exists
        if self.has_key(key):
            # This implementation deletes the object if non-versioned or adds a delete marker if versioned  # noqa: E501 # FIXME CoP
            s3.Object(self.bucket, s3_object_key).delete()
            return True
        else:
            return False

    def _has_key(self, key):  # type: ignore[explicit-override] # FIXME
        all_keys = self.list_keys()
        return key in all_keys

    def _assume_role_and_get_secret_credentials(self):
        role_session_name = "GXAssumeRoleSession"
        client = aws.boto3.client("sts", self._boto3_options.get("region_name"))
        role_arn = self._boto3_options.pop("assume_role_arn")
        assume_role_duration = self._boto3_options.pop("assume_role_duration")
        response = client.assume_role(
            RoleArn=role_arn,
            RoleSessionName=role_session_name,
            DurationSeconds=assume_role_duration,
        )
        self._boto3_options["aws_access_key_id"] = response["Credentials"]["AccessKeyId"]
        self._boto3_options["aws_secret_access_key"] = response["Credentials"]["SecretAccessKey"]
        self._boto3_options["aws_session_token"] = response["Credentials"]["SessionToken"]

    @property
    def boto3_options(self):
        result = {}
        if self._boto3_options.get("signature_version"):
            signature_version = self._boto3_options.pop("signature_version")
            result["config"] = aws.Config(signature_version=signature_version)
        if self._boto3_options.get("assume_role_arn"):
            self._assume_role_and_get_secret_credentials()
        result.update(self._boto3_options)

        return result

    def _create_client(self):
        return aws.boto3.client("s3", **self.boto3_options)

    def _create_resource(self):
        return aws.boto3.resource("s3", **self.boto3_options)

    @property
    @override
    def config(self) -> dict:
        return self._config


class TupleGCSStoreBackend(TupleStoreBackend):
    """
    Uses a GCS bucket as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.

    The filepath_template is a string template used to convert the key to a filepath.
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        bucket,
        project,
        prefix="",
        filepath_template=None,
        filepath_prefix=None,
        filepath_suffix=None,
        forbidden_substrings=None,
        platform_specific_separator=False,
        fixed_length_key=False,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        public_urls=True,
        base_public_path=None,
        store_name=None,
    ) -> None:
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            base_public_path=base_public_path,
            store_name=store_name,
        )
        self.bucket = bucket
        self.prefix = prefix
        self.project = project
        self._public_urls = public_urls
        # Initialize with store_backend_id if not part of an HTMLSiteStore
        if not self._suppress_store_backend_id:
            _ = self.store_backend_id

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "bucket": bucket,
            "project": project,
            "prefix": prefix,
            "filepath_template": filepath_template,
            "filepath_prefix": filepath_prefix,
            "filepath_suffix": filepath_suffix,
            "forbidden_substrings": forbidden_substrings,
            "platform_specific_separator": platform_specific_separator,
            "fixed_length_key": fixed_length_key,
            "suppress_store_backend_id": suppress_store_backend_id,
            "manually_initialize_store_backend_id": manually_initialize_store_backend_id,
            "public_urls": public_urls,
            "base_public_path": base_public_path,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def _build_gcs_object_key(self, key):
        if self.platform_specific_separator:
            if self.prefix:
                gcs_object_key = os.path.join(  # noqa: PTH118 # FIXME CoP
                    self.prefix, self._convert_key_to_filepath(key)
                )
            else:
                gcs_object_key = self._convert_key_to_filepath(key)
        else:  # noqa: PLR5501 # FIXME CoP
            if self.prefix:
                gcs_object_key = "/".join((self.prefix, self._convert_key_to_filepath(key)))
            else:
                gcs_object_key = self._convert_key_to_filepath(key)
        return gcs_object_key

    def _get(self, key):  # type: ignore[explicit-override] # FIXME
        from great_expectations.compatibility import google

        gcs = google.storage.Client(project=self.project)
        bucket = gcs.bucket(self.bucket)
        return self._get_by_gcs_object_key(bucket, key)

    @override
    def _get_all(self) -> list[Any]:
        from great_expectations.compatibility import google

        gcs = google.storage.Client(project=self.project)
        bucket = gcs.bucket(self.bucket)

        keys = self.list_keys()
        keys = [k for k in keys if k != StoreBackend.STORE_BACKEND_ID_KEY]

        return [self._get_by_gcs_object_key(bucket, key) for key in keys]

    def _get_by_gcs_object_key(self, bucket, key):
        gcs_object_key = self._build_gcs_object_key(key)
        gcs_response_object = bucket.get_blob(gcs_object_key)
        if not gcs_response_object:
            raise InvalidKeyError(  # noqa: TRY003 # FIXME CoP
                f"Unable to retrieve object from TupleGCSStoreBackend with the following Key: {key!s}"  # noqa: E501 # FIXME CoP
            )
        else:
            return gcs_response_object.download_as_bytes().decode("utf-8")

    def _set(  # type: ignore[explicit-override] # FIXME
        self,
        key,
        value,
        content_encoding="utf-8",
        content_type="application/json",
        **kwargs,
    ):
        gcs_object_key = self._build_gcs_object_key(key)

        from great_expectations.compatibility import google

        gcs = google.storage.Client(project=self.project)
        bucket = gcs.bucket(self.bucket)
        blob = bucket.blob(gcs_object_key)

        if isinstance(value, str):
            blob.content_encoding = content_encoding
            blob.upload_from_string(value.encode(content_encoding), content_type=content_type)
        else:
            blob.upload_from_string(value, content_type=content_type)
        return gcs_object_key

    @override
    def _move(self, source_key, dest_key, **kwargs) -> None:
        from great_expectations.compatibility import google

        gcs = google.storage.Client(project=self.project)
        bucket = gcs.bucket(self.bucket)

        source_filepath = self._convert_key_to_filepath(source_key)
        if not source_filepath.startswith(self.prefix):
            source_filepath = os.path.join(self.prefix, source_filepath)  # noqa: PTH118 # FIXME CoP
        dest_filepath = self._convert_key_to_filepath(dest_key)
        if not dest_filepath.startswith(self.prefix):
            dest_filepath = os.path.join(self.prefix, dest_filepath)  # noqa: PTH118 # FIXME CoP

        blob = bucket.blob(source_filepath)
        _ = bucket.rename_blob(blob, dest_filepath)

    @override
    def list_keys(self, prefix: Tuple = ()) -> List[Tuple]:
        # Note that the prefix arg is only included to maintain consistency with the parent class signature  # noqa: E501 # FIXME CoP
        key_list = []

        from great_expectations.compatibility import google

        gcs = google.storage.Client(self.project)

        for blob in gcs.list_blobs(self.bucket, prefix=self.prefix):
            gcs_object_name = blob.name
            gcs_object_key = os.path.relpath(
                gcs_object_name,
                self.prefix,
            )
            if self._is_missing_prefix_or_suffix(
                filepath_prefix=self.filepath_prefix,
                filepath_suffix=self.filepath_suffix,
                key=gcs_object_key,
            ):
                continue
            key = self._convert_filepath_to_key(gcs_object_key)
            if key:
                key_list.append(key)
        return key_list

    @override
    def get_url_for_key(self, key, protocol=None) -> str:
        path = self._convert_key_to_filepath(key)

        if self._public_urls:
            base_url = "https://storage.googleapis.com/"
        else:
            base_url = "https://storage.cloud.google.com/"

        path_url = self._get_path_url(path)

        return base_url + path_url

    def get_public_url_for_key(self, key, protocol=None):
        if not self.base_public_path:
            raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                """Error: No base_public_path was configured!
                    - A public URL was requested base_public_path was not configured for the
                """
            )
        path = self._convert_key_to_filepath(key)
        path_url = self._get_path_url(path)
        public_url = self.base_public_path + path_url
        return public_url

    def _get_path_url(self, path):
        if self.prefix:
            path_url = "/".join((self.bucket, self.prefix, path))
        else:  # noqa: PLR5501 # FIXME CoP
            if self.base_public_path:
                if self.base_public_path[-1] != "/":
                    path_url = f"/{path}"
                else:
                    path_url = path
            else:
                path_url = "/".join((self.bucket, path))
        return path_url

    def remove_key(self, key):  # type: ignore[explicit-override] # FIXME
        from great_expectations.compatibility import google

        gcs = google.storage.Client(project=self.project)
        bucket = gcs.bucket(self.bucket)
        try:
            bucket.delete_blobs(blobs=list(bucket.list_blobs(prefix=self.prefix)))
        except google.NotFound:
            return False
        return True

    def _has_key(self, key):  # type: ignore[explicit-override] # FIXME
        all_keys = self.list_keys()
        return key in all_keys


class TupleAzureBlobStoreBackend(TupleStoreBackend):
    """
    Uses an Azure Blob as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.
    The filepath_template is a string template used to convert the key to a filepath.

    You need to setup the connection string environment variable
    https://docs.microsoft.com/en-us/azure/storage/blobs/storage-quickstart-blobs-python
    """  # noqa: E501 # FIXME CoP

    # We will use blobclient here
    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        container,
        connection_string=None,
        credential=None,
        account_url=None,
        prefix=None,
        filepath_template=None,
        filepath_prefix=None,
        filepath_suffix=None,
        forbidden_substrings=None,
        platform_specific_separator=False,
        fixed_length_key=False,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        store_name=None,
    ) -> None:
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            store_name=store_name,
        )
        self.connection_string = connection_string or os.environ.get(  # noqa: TID251 # FIXME CoP
            "AZURE_STORAGE_CONNECTION_STRING"
        )
        self.credential = credential or os.environ.get(  # noqa: TID251 # FIXME CoP
            "AZURE_CREDENTIAL"
        )
        self.prefix = prefix or ""
        self.container = container
        self.account_url = account_url or os.environ.get(  # noqa: TID251 # FIXME CoP
            "AZURE_STORAGE_ACCOUNT_URL"
        )

    @property
    @functools.lru_cache  # noqa: B019 # lru_cache on method
    def _container_client(self) -> Any:
        from great_expectations.compatibility import azure

        # Validate that "azure" libraries were successfully imported and attempt to create "azure_client" handle.  # noqa: E501 # FIXME CoP
        if azure.BlobServiceClient:  # type: ignore[truthy-function] # False if NotImported
            try:
                if self.connection_string:
                    blob_service_client: azure.BlobServiceClient = (
                        azure.BlobServiceClient.from_connection_string(self.connection_string)
                    )
                elif self.account_url:
                    blob_service_client = azure.BlobServiceClient(
                        account_url=self.account_url,
                        credential=azure.DefaultAzureCredential(),
                    )
                elif self.credential and self.account_url:
                    blob_service_client = azure.BlobServiceClient(
                        account_url=self.account_url, credential=self.credential
                    )
                else:
                    raise StoreBackendError(  # noqa: TRY003, TRY301 # FIXME CoP
                        "Unable to initialize ServiceClient, AZURE_STORAGE_CONNECTION_STRING should be set"  # noqa: E501 # FIXME CoP
                    )
            except Exception as e:
                # Failure to create "azure_client" is most likely due invalid "azure_options" dictionary.  # noqa: E501 # FIXME CoP
                raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                    f'Due to exception: "{e!s}", "azure_client" could not be created.'
                ) from e
        else:
            raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                'Unable to create azure "BlobServiceClient" due to missing azure.storage.blob dependency.'  # noqa: E501 # FIXME CoP
            )

        return blob_service_client.get_container_client(self.container)

    def _get(self, key):  # type: ignore[explicit-override] # FIXME
        az_blob_key = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.prefix, self._convert_key_to_filepath(key)
        )
        return self._container_client.download_blob(az_blob_key).readall().decode("utf-8")

    @override
    def _get_all(self) -> list[Any]:
        keys = self.list_keys()
        return [self._get(key) for key in keys]

    def _set(self, key, value, content_encoding="utf-8", **kwargs):  # type: ignore[explicit-override] # FIXME
        from great_expectations.compatibility.azure import ContentSettings

        az_blob_key = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.prefix, self._convert_key_to_filepath(key)
        )

        if isinstance(value, str):
            if az_blob_key.endswith(".html"):
                my_content_settings = ContentSettings(content_type="text/html")
                self._container_client.upload_blob(
                    name=az_blob_key,
                    data=value,
                    encoding=content_encoding,
                    overwrite=True,
                    content_settings=my_content_settings,
                )
            else:
                self._container_client.upload_blob(
                    name=az_blob_key,
                    data=value,
                    encoding=content_encoding,
                    overwrite=True,
                )
        else:
            self._container_client.upload_blob(name=az_blob_key, data=value, overwrite=True)
        return az_blob_key

    @override
    def list_keys(self, prefix: Tuple = ()) -> List[Tuple]:
        # Note that the prefix arg is only included to maintain consistency with the parent class signature  # noqa: E501 # FIXME CoP
        key_list = []

        for obj in self._container_client.list_blobs(name_starts_with=self.prefix):
            az_blob_key = os.path.relpath(obj.name)
            if az_blob_key.startswith(f"{self.prefix}{os.path.sep}"):
                az_blob_key = az_blob_key[len(self.prefix) + 1 :]

            if self._is_missing_prefix_or_suffix(
                filepath_prefix=self.filepath_prefix,
                filepath_suffix=self.filepath_suffix,
                key=az_blob_key,
            ):
                continue
            key = self._convert_filepath_to_key(az_blob_key)

            key_list.append(key)
        return key_list

    @override
    def get_url_for_key(self, key, protocol=None) -> str:
        az_blob_key = self._convert_key_to_filepath(key)
        az_blob_path = os.path.join(  # noqa: PTH118 # FIXME CoP
            self.container, self.prefix, az_blob_key
        )

        return f"https://{self._container_client.account_name}.blob.core.windows.net/{az_blob_path}"

    def _has_key(self, key):  # type: ignore[explicit-override] # FIXME
        all_keys = self.list_keys()
        return key in all_keys

    @override
    def _move(self, source_key, dest_key, **kwargs) -> None:
        source_blob_path = self._convert_key_to_filepath(source_key)
        if not source_blob_path.startswith(self.prefix):
            source_blob_path = os.path.join(  # noqa: PTH118 # FIXME CoP
                self.prefix, source_blob_path
            )
        dest_blob_path = self._convert_key_to_filepath(dest_key)
        if not dest_blob_path.startswith(self.prefix):
            dest_blob_path = os.path.join(self.prefix, dest_blob_path)  # noqa: PTH118 # FIXME CoP

        # azure storage sdk does not have _move method
        source_blob = self._container_client.get_blob_client(source_blob_path)
        dest_blob = self._container_client.get_blob_client(dest_blob_path)

        dest_blob.start_copy_from_url(source_blob.url, requires_sync=True)
        copy_properties = dest_blob.get_blob_properties().copy

        if copy_properties.status != "success":
            dest_blob.abort_copy(copy_properties.id)
            raise StoreBackendError(  # noqa: TRY003 # FIXME CoP
                f"Unable to copy blob {source_blob_path} with status {copy_properties.status}"
            )
        source_blob.delete_blob()

    def remove_key(self, key):  # type: ignore[explicit-override] # FIXME
        if not isinstance(key, tuple):
            key = key.to_tuple()

        az_blob_path = self._convert_key_to_filepath(key)
        if not az_blob_path.startswith(self.prefix):
            az_blob_path = os.path.join(self.prefix, az_blob_path)  # noqa: PTH118 # FIXME CoP

        blob = self._container_client.get_blob_client(az_blob_path)
        blob.delete_blob()
        return True

    @property
    @override
    def config(self) -> dict:
        return self._config  # type: ignore[attr-defined] # FIXME CoP



import pandas as pd

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.values_not_match_regex_values import (
    ColumnValuesNotMatchRegexValues,
    ColumnValuesNotMatchRegexValuesResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import SQL_DATA_SOURCES, SnowflakeDatasourceTestConfig

COLUMN_NAME = "whatevs"
BIG_NUMBER = 101
MATCH_NONE_REGEX = "^$"  # Regex that matches nothing

DATA_FRAME = pd.DataFrame({COLUMN_NAME: ["abc", "def", "ghi", "1ab2", None]})
DATA_FRAME_WITH_LOTS_OF_VALUES = pd.DataFrame({COLUMN_NAME: ["A"] * BIG_NUMBER})

SQL_DATA_SOURCES_EXCEPT_SNOWFLAKE = [
    datasource
    for datasource in SQL_DATA_SOURCES
    if not isinstance(datasource, SnowflakeDatasourceTestConfig)
]


class TestColumnValuesNotMatchRegexValues:
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES_EXCEPT_SNOWFLAKE,
        data=DATA_FRAME,
    )
    def test_partial_match_characters(self, batch_for_datasource: Batch) -> None:
        metric = ColumnValuesNotMatchRegexValues(column=COLUMN_NAME, regex="ab")
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexValuesResult)
        # Expect values that DO NOT contain 'ab'
        assert sorted(metric_result.value) == ["def", "ghi"]

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_special_characters(self, batch_for_datasource: Batch) -> None:
        metric = ColumnValuesNotMatchRegexValues(column=COLUMN_NAME, regex="^(a|d).+")
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexValuesResult)
        # Expect values that DO NOT start with 'a' or 'd'
        assert sorted(metric_result.value) == ["1ab2", "ghi"]

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME_WITH_LOTS_OF_VALUES,
    )
    def test_default_limit(self, batch_for_datasource: Batch) -> None:
        # Use a regex that matches nothing to get all values back
        metric = ColumnValuesNotMatchRegexValues(column=COLUMN_NAME, regex=MATCH_NONE_REGEX)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexValuesResult)
        # Should return up to the default limit (20)
        assert len(metric_result.value) == 20
        assert all(val == "A" for val in metric_result.value)

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME_WITH_LOTS_OF_VALUES,
    )
    def test_custom_limit(self, batch_for_datasource: Batch) -> None:
        limit = 7
        # Use a regex that matches nothing to get all values back
        metric = ColumnValuesNotMatchRegexValues(
            column=COLUMN_NAME, regex=MATCH_NONE_REGEX, limit=limit
        )
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnValuesNotMatchRegexValuesResult)
        assert len(metric_result.value) == limit
        assert all(val == "A" for val in metric_result.value)



from __future__ import annotations

from typing import TYPE_CHECKING

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    column_aggregate_partial,
    column_aggregate_value,
)

if TYPE_CHECKING:
    import pandas as pd


class ColumnValuesLengthMax(ColumnAggregateMetricProvider):
    metric_name = "column_values.length.max"

    @column_aggregate_value(engine=PandasExecutionEngine, filter_column_isnull=True)  # type: ignore[misc] # untyped-decorator
    def _pandas(cls, column: pd.Series, **kwargs: dict) -> int:
        return column.map(len).max()

    @column_aggregate_partial(  # type: ignore[misc] # untyped-decorator
        engine=SqlAlchemyExecutionEngine, filter_column_isnull=True
    )
    def _sqlalchemy(cls, column, **kwargs: dict):
        return sa.func.max(sa.func.length(column))

    @column_aggregate_partial(engine=SparkDFExecutionEngine, filter_column_isnull=True)  # type: ignore[misc] # untyped-decorator
    def _spark(cls, column, **kwargs: dict):
        return F.max(F.length(column))



from __future__ import annotations

import json
import logging
import pathlib
from collections.abc import MutableMapping, MutableSequence
from io import StringIO
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    Mapping,
    Type,
    Union,
    overload,
)

from ruamel.yaml import YAML

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.constants import (
    _ASSETS_KEY,
    _FIELDS_ALWAYS_SET,
)

if TYPE_CHECKING:
    MappingIntStrAny = Mapping[Union[int, str], Any]
    AbstractSetIntStr = AbstractSet[Union[int, str]]
    from typing_extensions import Self

    from great_expectations.core.config_provider import _ConfigurationProvider

logger = logging.getLogger(__name__)

yaml = YAML(typ="safe")
# NOTE (kilo59): the following settings appear to be what we use in existing codebase
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False


class FluentBaseModel(pydantic.BaseModel):
    """
    Base model for most fluent datasource related pydantic models.

    Adds yaml dumping and parsing methods.

    Extra fields are not allowed.

    Serialization methods default to `exclude_unset = True` to prevent serializing
    configs full of mostly unset default values.
    Also prevents passing along unset kwargs to BatchSpec.
    https://docs.pydantic.dev/usage/exporting_models/
    """

    # Due to namespace collisions with certain keywords like 'schema', we've set the default of
    # `by_alias` for the various serialization methods to `True`.
    # If we're using an alias, the assumption is that we want to serialize with that alias.
    # Related FastAPI thread that discusses overriding this default: https://github.com/tiangolo/fastapi/discussions/2753

    class Config:
        extra = pydantic.Extra.forbid

    @classmethod
    def parse_yaml(cls: Type[Self], f: Union[pathlib.Path, str]) -> Self:
        loaded = yaml.load(f)
        logger.debug(f"loaded from yaml ->\n{pf(loaded, depth=3)}\n")
        # noinspection PyArgumentList
        config = cls(**loaded)
        return config

    @overload
    def yaml(
        self,
        stream_or_path: Union[StringIO, None] = None,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **yaml_kwargs,
    ) -> str: ...

    @overload
    def yaml(
        self,
        stream_or_path: pathlib.Path,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **yaml_kwargs,
    ) -> pathlib.Path: ...

    def yaml(  # noqa: PLR0913 # FIXME CoP
        self,
        stream_or_path: Union[StringIO, pathlib.Path, None] = None,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        by_alias: bool = True,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Union[Callable[[Any], Any], None] = None,
        models_as_dict: bool = True,
        **yaml_kwargs,
    ) -> Union[str, pathlib.Path]:
        """
        Serialize the config object as yaml.

        Writes to a file if a `pathlib.Path` is provided.
        Else it writes to a stream and returns a yaml string.
        """
        if stream_or_path is None:
            stream_or_path = StringIO()

        # pydantic json encoder has support for many more types
        # TODO: can we dump json string directly to yaml.dump?
        intermediate_json = self._json_dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
        )
        yaml.dump(intermediate_json, stream=stream_or_path, **yaml_kwargs)

        if isinstance(stream_or_path, pathlib.Path):
            return stream_or_path
        return stream_or_path.getvalue()

    @override
    def json(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = True,
        # deprecated - use exclude_unset instead
        skip_defaults: bool | None = None,
        # Default to True to prevent serializing long configs full of unset default values
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Callable[[Any], Any] | None = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        """
        Generate a JSON representation of the model, `include` and `exclude` arguments
        as per `dict()`.

        `encoder` is an optional function to supply as `default` to json.dumps(), other
        arguments as per `json.dumps()`.

        Deviates from pydantic `exclude_unset` `True` by default instead of `False` by
        default.
        """
        self.__fields_set__.update(_FIELDS_ALWAYS_SET)
        _update__fields_set__on_truthyness(self, _ASSETS_KEY)

        return super().json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
            **dumps_kwargs,
        )

    def _json_dict(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        by_alias: bool = True,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Union[Callable[[Any], Any], None] = None,
        models_as_dict: bool = True,
        **dumps_kwargs,
    ) -> dict:
        """
        JSON compatible dictionary. All complex types removed.
        Prefer `.dict()` or `.json()`
        """
        return json.loads(
            self.json(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                encoder=encoder,
                models_as_dict=models_as_dict,
                **dumps_kwargs,
            )
        )

    @override
    def dict(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = True,
        # Default to True to prevent serializing long configs full of unset default values
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        # deprecated - use exclude_unset instead
        skip_defaults: bool | None = None,
        # custom
        config_provider: _ConfigurationProvider | None = None,
        raise_on_missing_config_provider: bool = False,
    ) -> dict[str, Any]:
        """
        Generate a dictionary representation of the model, optionally specifying which
        fields to include or exclude.

        Deviates from pydantic `exclude_unset` `True` by default instead of `False` by
        default.
        """
        self.__fields_set__.update(_FIELDS_ALWAYS_SET)
        _update__fields_set__on_truthyness(self, _ASSETS_KEY)

        result = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            skip_defaults=skip_defaults,
        )

        class_name = self.__class__.__name__
        if config_provider:
            logger.debug(f"{class_name}.dict() - substituting config values")
            _recursively_set_config_value(result, config_provider)
        elif raise_on_missing_config_provider:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"{class_name}.dict() -"
                " `config_provider` must be provided if `raise_on_missing_config_provider` is True."
                f" {class_name} may be missing a context."
            )
        else:
            logger.info(
                f"{class_name}.dict() - missing `config_provider`, skipping config substitution"
            )

        return result

    @staticmethod
    def _include_exclude_to_dict(
        include_exclude: AbstractSetIntStr | MappingIntStrAny | None,
    ) -> Dict[int | str, Any]:
        """
        Takes the mapping or abstract set passed to pydantic model export include or exclude and makes it a
        mutable dictionary that can be altered for nested include/exclude operations.

        See: https://docs.pydantic.dev/usage/exporting_models/#advanced-include-and-exclude

        Args:
            include_exclude: The include or exclude key passed to pydantic model export methods.

        Returns: A mutable dictionary that can be used for nested include/exclude.
        """  # noqa: E501 # FIXME CoP
        if isinstance(include_exclude, Mapping):
            include_exclude_dict = dict(include_exclude)
        elif isinstance(include_exclude, AbstractSet):
            include_exclude_dict = dict.fromkeys(include_exclude, True)
        else:
            include_exclude_dict = {}
        return include_exclude_dict

    @override
    def __str__(self):
        return self.yaml()


class GenericBaseModel(FluentBaseModel, pydantic.GenericModel): ...


def _recursively_set_config_value(  # noqa: C901 #  too complex
    data: MutableMapping | MutableSequence, config_provider: _ConfigurationProvider
) -> None:
    if isinstance(data, MutableMapping):
        for k, v in data.items():
            if isinstance(v, ConfigStr):
                data[k] = v.get_config_value(config_provider)
            elif isinstance(v, (MutableMapping, MutableSequence)):
                _recursively_set_config_value(v, config_provider)
    elif isinstance(data, MutableSequence):
        for i, v in enumerate(data):
            if isinstance(v, ConfigStr):
                data[i] = v.get_config_value(config_provider)
            elif isinstance(v, (MutableMapping, MutableSequence)):
                _recursively_set_config_value(v, config_provider)


def _update__fields_set__on_truthyness(model: FluentBaseModel, field_name: str) -> None:
    """
    This method updates the special `__fields__set__` attribute if the provided field is
    present and the value truthy. Otherwise it removes the entry from `__fields_set__`.

    For background `__fields_set__` is what determines whether or not a field is
    serialized when `exclude_unset` is used with `.dict()`/`.json()`/`.yaml()`.

    This is set automatically in most cases, but if a field was set with a `pre`
    validator then this will not have been updated and so if we want it to be dumped
    when `exclude_unset` is used we need to update `__fields_set__`.

    https://docs.pydantic.dev/usage/validators/#pre-and-per-item-validators
    https://docs.pydantic.dev/usage/exporting_models/#modeldict
    """
    if getattr(model, field_name, None):
        model.__fields_set__.add(field_name)
        logger.debug(f"{model.__class__.__name__}.__fields_set__ {field_name} added")
    else:
        model.__fields_set__.discard(field_name)
        logger.debug(f"{model.__class__.__name__}.__fields_set__ {field_name} discarded")



from __future__ import annotations

import inspect
import logging
import warnings
from typing import TYPE_CHECKING

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import (
    MetricFunctionTypes,
    MetricPartialFunctionTypes,
    MetricPartialFunctionTypeSuffixes,
    SummarizationMetricNameSuffixes,
)
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider.column_map_condition_auxilliary_methods import (  # noqa: E501 # FIXME CoP
    _pandas_column_map_condition_value_counts,
    _pandas_column_map_condition_values,
    _spark_column_map_condition_value_counts,
    _spark_column_map_condition_values,
    _sqlalchemy_column_map_condition_value_counts,
    _sqlalchemy_column_map_condition_values,
)
from great_expectations.expectations.metrics.map_metric_provider.column_pair_map_condition_auxilliary_methods import (  # noqa: E501 # FIXME CoP
    _pandas_column_pair_map_condition_filtered_row_count,
    _pandas_column_pair_map_condition_values,
    _spark_column_pair_map_condition_filtered_row_count,
    _spark_column_pair_map_condition_values,
    _sqlalchemy_column_pair_map_condition_filtered_row_count,
    _sqlalchemy_column_pair_map_condition_values,
)
from great_expectations.expectations.metrics.map_metric_provider.is_sqlalchemy_metric_selectable import (  # noqa: E501 # FIXME CoP
    _is_sqlalchemy_metric_selectable,
)
from great_expectations.expectations.metrics.map_metric_provider.map_condition_auxilliary_methods import (  # noqa: E501 # FIXME CoP
    _pandas_map_condition_index,
    _pandas_map_condition_query,
    _pandas_map_condition_rows,
    _pandas_map_condition_unexpected_count,
    _spark_map_condition_index,
    _spark_map_condition_query,
    _spark_map_condition_rows,
    _spark_map_condition_unexpected_count_aggregate_fn,
    _spark_map_condition_unexpected_count_value,
    _sqlalchemy_map_condition_index,
    _sqlalchemy_map_condition_query,
    _sqlalchemy_map_condition_rows,
    _sqlalchemy_map_condition_unexpected_count_aggregate_fn,
    _sqlalchemy_map_condition_unexpected_count_value,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_map_condition_auxilliary_methods import (  # noqa: E501 # FIXME CoP
    _pandas_multicolumn_map_condition_filtered_row_count,
    _pandas_multicolumn_map_condition_values,
    _spark_multicolumn_map_condition_filtered_row_count,
    _spark_multicolumn_map_condition_values,
    _sqlalchemy_multicolumn_map_condition_filtered_row_count,
    _sqlalchemy_multicolumn_map_condition_values,
)
from great_expectations.expectations.metrics.metric_provider import (
    MetricProvider,
)
from great_expectations.expectations.registry import (
    get_metric_provider,
    register_metric,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.expectations.metrics import MetaMetricProvider

logger = logging.getLogger(__name__)


class MapMetricProvider(MetricProvider):
    """The base class for defining metrics that are evaluated for every row. An example of a map metric is
    `column_values.null` (which is implemented as a `ColumnMapMetricProvider`, a subclass of `MapMetricProvider`).
    """  # noqa: E501 # FIXME CoP

    condition_domain_keys: tuple[str, ...] = (
        "batch_id",
        "table",
        "row_condition",
        "condition_parser",
    )
    function_domain_keys: tuple[str, ...] = (
        "batch_id",
        "table",
        "row_condition",
        "condition_parser",
    )
    condition_value_keys: tuple[str, ...] = tuple()
    function_value_keys: tuple[str, ...] = tuple()
    filter_column_isnull = True

    @classmethod
    def _register_metric_functions(cls):  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        if not (hasattr(cls, "function_metric_name") or hasattr(cls, "condition_metric_name")):
            return

        for attr, candidate_metric_fn in inspect.getmembers(cls):
            if not hasattr(candidate_metric_fn, "metric_engine"):
                # This is not a metric.
                continue

            metric_fn_type = getattr(candidate_metric_fn, "metric_fn_type", None)
            if not metric_fn_type:
                # This is not a metric (valid metrics possess exectly one metric function).
                return

            engine = candidate_metric_fn.metric_engine
            if not issubclass(engine, ExecutionEngine):
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "Metric functions must be defined with an ExecutionEngine as part of registration."  # noqa: E501 # FIXME CoP
                )

            if metric_fn_type in [
                MetricPartialFunctionTypes.MAP_CONDITION_FN,
                MetricPartialFunctionTypes.MAP_CONDITION_SERIES,
                MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
            ]:
                if not hasattr(cls, "condition_metric_name"):
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        """A "MapMetricProvider" must have a "condition_metric_name" to have a decorated \
"column_condition_partial" method."""  # noqa: E501 # FIXME CoP
                    )

                condition_provider = candidate_metric_fn
                # noinspection PyUnresolvedReferences
                metric_name = cls.condition_metric_name
                metric_domain_keys = cls.condition_domain_keys
                metric_value_keys = cls.condition_value_keys
                metric_definition_kwargs = getattr(
                    condition_provider, "metric_definition_kwargs", {}
                )
                domain_type = getattr(
                    condition_provider,
                    "domain_type",
                    metric_definition_kwargs.get("domain_type", MetricDomainTypes.TABLE),
                )
                if issubclass(engine, PandasExecutionEngine):
                    register_metric(
                        metric_name=f"{metric_name}.{metric_fn_type.metric_suffix}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=condition_provider,
                        metric_fn_type=metric_fn_type,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_pandas_map_condition_unexpected_count,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_pandas_map_condition_index,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_pandas_map_condition_query,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_ROWS.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_pandas_map_condition_rows,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    if domain_type == MetricDomainTypes.COLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_column_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUE_COUNTS.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_column_map_condition_value_counts,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.COLUMN_PAIR:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_column_pair_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_column_pair_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.MULTICOLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_multicolumn_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_pandas_multicolumn_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                elif issubclass(engine, SqlAlchemyExecutionEngine):
                    register_metric(
                        metric_name=f"{metric_name}.{metric_fn_type.metric_suffix}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=condition_provider,
                        metric_fn_type=metric_fn_type,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_ROWS.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_sqlalchemy_map_condition_rows,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_sqlalchemy_map_condition_index,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_sqlalchemy_map_condition_query,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    if metric_fn_type == MetricPartialFunctionTypes.MAP_CONDITION_FN:
                        # Documentation in "MetricProvider._register_metric_functions()" explains registration protocol.  # noqa: E501 # FIXME CoP
                        if domain_type == MetricDomainTypes.COLUMN:
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=_sqlalchemy_map_condition_unexpected_count_aggregate_fn,
                                metric_fn_type=MetricPartialFunctionTypes.AGGREGATE_FN,
                            )
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=None,
                                metric_fn_type=MetricFunctionTypes.VALUE,
                            )
                        else:
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=_sqlalchemy_map_condition_unexpected_count_value,
                                metric_fn_type=MetricFunctionTypes.VALUE,
                            )
                    elif metric_fn_type == MetricPartialFunctionTypes.WINDOW_CONDITION_FN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=metric_value_keys,
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_map_condition_unexpected_count_value,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    if domain_type == MetricDomainTypes.COLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_column_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUE_COUNTS.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_column_map_condition_value_counts,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.COLUMN_PAIR:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_column_pair_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_column_pair_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.MULTICOLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_multicolumn_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_sqlalchemy_multicolumn_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                elif issubclass(engine, SparkDFExecutionEngine):
                    register_metric(
                        metric_name=f"{metric_name}.{metric_fn_type.metric_suffix}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=condition_provider,
                        metric_fn_type=metric_fn_type,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_ROWS.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_spark_map_condition_rows,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_spark_map_condition_index,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    register_metric(
                        metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=(*metric_value_keys, "result_format"),
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=_spark_map_condition_query,
                        metric_fn_type=MetricFunctionTypes.VALUE,
                    )
                    if metric_fn_type == MetricPartialFunctionTypes.MAP_CONDITION_FN:
                        # Documentation in "MetricProvider._register_metric_functions()" explains registration protocol.  # noqa: E501 # FIXME CoP
                        if domain_type == MetricDomainTypes.COLUMN:
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=_spark_map_condition_unexpected_count_aggregate_fn,
                                metric_fn_type=MetricPartialFunctionTypes.AGGREGATE_FN,
                            )
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=None,
                                metric_fn_type=MetricFunctionTypes.VALUE,
                            )
                        else:
                            register_metric(
                                metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                                metric_domain_keys=metric_domain_keys,
                                metric_value_keys=metric_value_keys,
                                execution_engine=engine,
                                metric_class=cls,
                                metric_provider=_spark_map_condition_unexpected_count_value,
                                metric_fn_type=MetricFunctionTypes.VALUE,
                            )
                    elif metric_fn_type == MetricPartialFunctionTypes.WINDOW_CONDITION_FN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=metric_value_keys,
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_map_condition_unexpected_count_value,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    if domain_type == MetricDomainTypes.COLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_column_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUE_COUNTS.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_column_map_condition_value_counts,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.COLUMN_PAIR:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_column_pair_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_column_pair_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                    elif domain_type == MetricDomainTypes.MULTICOLUMN:
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_multicolumn_map_condition_values,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
                        register_metric(
                            metric_name=f"{metric_name}.{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                            metric_domain_keys=metric_domain_keys,
                            metric_value_keys=(*metric_value_keys, "result_format"),
                            execution_engine=engine,
                            metric_class=cls,
                            metric_provider=_spark_multicolumn_map_condition_filtered_row_count,
                            metric_fn_type=MetricFunctionTypes.VALUE,
                        )
            elif metric_fn_type in [
                MetricPartialFunctionTypes.MAP_FN,
                MetricPartialFunctionTypes.MAP_SERIES,
                MetricPartialFunctionTypes.WINDOW_FN,
            ]:
                if not hasattr(cls, "function_metric_name"):
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        """A "MapMetricProvider" must have a "function_metric_name" to have a decorated \
"column_function_partial" method."""  # noqa: E501 # FIXME CoP
                    )

                map_function_provider = candidate_metric_fn
                # noinspection PyUnresolvedReferences
                metric_name = cls.function_metric_name
                metric_domain_keys = cls.function_domain_keys
                metric_value_keys = cls.function_value_keys
                register_metric(
                    metric_name=f"{metric_name}.{metric_fn_type.metric_suffix}",
                    metric_domain_keys=metric_domain_keys,
                    metric_value_keys=metric_value_keys,
                    execution_engine=engine,
                    metric_class=cls,
                    metric_provider=map_function_provider,
                    metric_fn_type=metric_fn_type,
                )

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: ExpectationConfiguration | None = None,
        execution_engine: ExecutionEngine | None = None,
        runtime_configuration: dict | None = None,
    ):
        dependencies: dict[str, MetricConfiguration] = {}

        base_metric_value_kwargs = {
            k: v for k, v in metric.metric_value_kwargs.items() if k != "result_format"
        }

        metric_name: str = metric.metric_name

        metric_suffix: str = f".{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"

        # Documentation in "MetricProvider._register_metric_functions()" explains registration/dependency protocol.  # noqa: E501 # FIXME CoP
        if metric_name.endswith(metric_suffix):
            has_aggregate_fn: bool = False

            if execution_engine is not None:
                try:
                    _ = get_metric_provider(
                        f"{metric_name}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                        execution_engine,
                    )
                    has_aggregate_fn = True
                except gx_exceptions.MetricProviderError:
                    pass

            if has_aggregate_fn:
                dependencies["metric_partial_fn"] = MetricConfiguration(
                    metric_name=f"{metric_name}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                    metric_domain_kwargs=metric.metric_domain_kwargs,
                    metric_value_kwargs=base_metric_value_kwargs,
                )
            else:
                dependencies["unexpected_condition"] = MetricConfiguration(
                    metric_name=(
                        f"{metric_name[: -len(metric_suffix)]}."
                        f"{MetricPartialFunctionTypeSuffixes.CONDITION.value}"
                    ),
                    metric_domain_kwargs=metric.metric_domain_kwargs,
                    metric_value_kwargs=base_metric_value_kwargs,
                )

        # MapMetric uses "condition" metric to build "unexpected_count.aggregate_fn" and other listed metrics as well.  # noqa: E501 # FIXME CoP
        unexpected_condition_dependent_metric_name_suffixes: list[str] = list(
            filter(
                lambda element: metric_name.endswith(element),
                [
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_VALUE_COUNTS.value}",
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
                    f".{SummarizationMetricNameSuffixes.FILTERED_ROW_COUNT.value}",
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}",
                    f".{SummarizationMetricNameSuffixes.UNEXPECTED_ROWS.value}",
                ],
            )
        )
        if len(unexpected_condition_dependent_metric_name_suffixes) == 1:
            metric_suffix = unexpected_condition_dependent_metric_name_suffixes[0]
            if metric_name.endswith(metric_suffix):
                dependencies["unexpected_condition"] = MetricConfiguration(
                    metric_name=f"{metric_name[: -len(metric_suffix)]}."
                    f"{MetricPartialFunctionTypeSuffixes.CONDITION.value}",
                    metric_domain_kwargs=metric.metric_domain_kwargs,
                    metric_value_kwargs=base_metric_value_kwargs,
                )

        return dependencies

    @staticmethod
    def is_sqlalchemy_metric_selectable(
        map_metric_provider: MetaMetricProvider,
    ) -> bool:
        # deprecated-v0.16.1
        warnings.warn(
            "MapMetricProvider.is_sqlalchemy_metric_selectable is deprecated."
            "You can use the great_expectations.expectations.metrics.map_metric_provider.is_sqlalchemy_metric_selectable._is_sqlalchemy_metric_selectable function, but please note that it is not considered part of the public API, and could change in the future.",  # noqa: E501 # FIXME CoP
            DeprecationWarning,
        )

        return _is_sqlalchemy_metric_selectable(map_metric_provider)



from logging import Logger
from typing import Any, ClassVar, Literal, Optional, Type

from great_expectations.compatibility import azure
from great_expectations.datasource.fluent import _SparkFilePathDatasource
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset
from great_expectations.datasource.fluent.data_connector import (
    S3DataConnector,
)
from great_expectations.datasource.fluent.interfaces import (
    BatchMetadata,
)
from great_expectations.datasource.fluent.spark_datasource import (
    SparkDatasourceError,
)

logger: Logger

class SparkAzureBlobStorageDatasourceError(SparkDatasourceError): ...

class SparkAzureBlobStorageDatasource(_SparkFilePathDatasource):
    # class attributes
    data_connector_type: ClassVar[Type[S3DataConnector]] = ...

    # instance attributes
    type: Literal["spark_abs"] = "spark_abs"

    # Azure Blob Storage specific attributes
    azure_options: dict[str, ConfigStr | Any] = {}
    # private
    _azure_client: azure.BlobServiceClient | None
    def add_csv_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        abs_container: str = ...,
        abs_name_starts_with: str = "",
        abs_delimiter: str = "/",
        abs_recursive_file_discovery: bool = False,
        header: bool = ...,
        infer_schema: bool = ...,
    ) -> CSVAsset: ...



from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    NON_SQL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import MySQLDatasourceTestConfig

UNIQUE_INTS = "unique_integers"
DUPLICATE_INTS = "duplicate_integers"
UNIQUE_STRINGS = "unique_strings"
DUPLICATE_STRINGS = "duplicate_strings"
UNIQUE_WITH_NULL = "unique_with_null"
DUPLICATE_WITH_NULL = "duplicate_with_null"

DATA = pd.DataFrame(
    {
        UNIQUE_INTS: [1, 2, 3, 4],
        DUPLICATE_INTS: [1, 2, 3, 3],
        UNIQUE_STRINGS: ["a", "b", "c", "d"],
        DUPLICATE_STRINGS: ["a", "b", "c", "c"],
        UNIQUE_WITH_NULL: [1, 2, None, None],
        DUPLICATE_WITH_NULL: [1, 1, None, None],
    },
    dtype="object",
)

SUPPORTED_SQL_DATASOURCES = [
    ds
    for ds in SQL_DATA_SOURCES
    if not isinstance(ds, MySQLDatasourceTestConfig)  # why don't we support MySQL?
]


@parameterize_batch_for_data_sources(data_source_configs=NON_SQL_DATA_SOURCES, data=DATA)
def test_success_complete_non_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeUnique(
        column=UNIQUE_INTS,
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_SQL_DATASOURCES, data=DATA)
def test_success_complete_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeUnique(
        column=UNIQUE_INTS,
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 4,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "missing_count": 0,
        "missing_percent": 0.0,
        "unexpected_percent_total": 0.0,
        "unexpected_percent_nonmissing": 0.0,
        "unexpected_index_query": ANY,
        "partial_unexpected_counts": [],
        "unexpected_list": [],
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=UNIQUE_INTS,
            ),
            id="unique_integers",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=UNIQUE_STRINGS,
            ),
            id="unique_strings",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=UNIQUE_WITH_NULL,
            ),
            id="unique_with_null",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(column=DUPLICATE_INTS, mostly=0.5),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeUnique,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=DUPLICATE_INTS,
            ),
            id="duplicate_integers",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=DUPLICATE_STRINGS,
            ),
            id="duplicate_strings",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(
                column=DUPLICATE_WITH_NULL,
            ),
            id="duplicate_nulls",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeUnique(column=DUPLICATE_INTS, mostly=0.7),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeUnique,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from __future__ import annotations

import pytest

from great_expectations.core.domain import SemanticDomainTypes
from great_expectations.experimental.metric_repository.column_filter import ColumnFilter
from great_expectations.validator.validator import Validator


@pytest.fixture
def mock_validator(mocker):
    """Create a mock validator for testing."""
    validator = mocker.Mock(spec=Validator)
    validator.active_batch_id = "test_batch_id"
    return validator


@pytest.fixture
def sample_column_names():
    """Sample column names for testing."""
    return ["id", "name", "age", "salary", "created_at", "is_active", "description"]


@pytest.fixture
def sample_column_types():
    """Sample column type information for testing."""
    return [
        {"name": "id", "type": "INTEGER"},
        {"name": "name", "type": "VARCHAR"},
        {"name": "age", "type": "INTEGER"},
        {"name": "salary", "type": "DECIMAL"},
        {"name": "created_at", "type": "TIMESTAMP"},
        {"name": "is_active", "type": "BOOLEAN"},
        {"name": "description", "type": "TEXT"},
    ]


class TestColumnFilter:
    """Test cases for ColumnFilter class."""

    @pytest.mark.unit
    def test_init_with_defaults(self):
        """Test initialization with default parameters."""
        column_filter = ColumnFilter()
        assert column_filter._include_column_names == []
        assert column_filter._exclude_column_names == []
        assert column_filter._include_column_name_suffixes == []
        assert column_filter._exclude_column_name_suffixes == []
        assert column_filter._include_semantic_types == []
        assert column_filter._exclude_semantic_types == []

    @pytest.mark.unit
    def test_init_with_parameters(self):
        """Test initialization with specific parameters."""
        column_filter = ColumnFilter(
            include_column_names=["col1", "col2"],
            exclude_column_names=["col3"],
            include_semantic_types=[SemanticDomainTypes.NUMERIC],
            exclude_semantic_types=[SemanticDomainTypes.TEXT],
        )
        assert column_filter._include_column_names == ["col1", "col2"]
        assert column_filter._exclude_column_names == ["col3"]
        assert column_filter._include_semantic_types == [SemanticDomainTypes.NUMERIC]
        assert column_filter._exclude_semantic_types == [SemanticDomainTypes.TEXT]

    @pytest.mark.unit
    def test_normalize_semantic_types_none(self):
        """Test semantic type normalization with None."""
        column_filter = ColumnFilter()
        result = column_filter._normalize_semantic_types(None)
        assert result == []

    @pytest.mark.unit
    def test_normalize_semantic_types_single(self):
        """Test semantic type normalization with single type."""
        column_filter = ColumnFilter()
        result = column_filter._normalize_semantic_types(SemanticDomainTypes.NUMERIC)
        assert result == [SemanticDomainTypes.NUMERIC]

    @pytest.mark.unit
    def test_normalize_semantic_types_list(self):
        """Test semantic type normalization with list."""
        column_filter = ColumnFilter()
        types = [SemanticDomainTypes.NUMERIC, SemanticDomainTypes.TEXT]
        result = column_filter._normalize_semantic_types(types)
        assert result == types

    @pytest.mark.unit
    def test_get_table_column_names(self, mock_validator, sample_column_names):
        """Test getting table column names."""
        mock_validator.get_metric.return_value = sample_column_names

        column_filter = ColumnFilter()
        result = column_filter._get_table_column_names(mock_validator)

        assert result == sample_column_names
        mock_validator.get_metric.assert_called_once()

    @pytest.mark.unit
    def test_apply_column_name_filters_no_filters(self, sample_column_names):
        """Test column name filtering with no filters applied."""
        column_filter = ColumnFilter()
        result = column_filter._apply_column_name_filters(sample_column_names)
        assert result == sample_column_names

    @pytest.mark.unit
    def test_apply_column_name_filters_include_names(self, sample_column_names):
        """Test column name filtering with include names."""
        column_filter = ColumnFilter(include_column_names=["id", "name"])
        result = column_filter._apply_column_name_filters(sample_column_names)
        assert result == ["id", "name"]

    @pytest.mark.unit
    def test_apply_column_name_filters_exclude_names(self, sample_column_names):
        """Test column name filtering with exclude names."""
        column_filter = ColumnFilter(exclude_column_names=["id", "description"])
        result = column_filter._apply_column_name_filters(sample_column_names)
        expected = ["name", "age", "salary", "created_at", "is_active"]
        assert result == expected

    @pytest.mark.unit
    def test_apply_column_name_filters_include_suffixes(self, sample_column_names):
        """Test column name filtering with include suffixes."""
        column_filter = ColumnFilter(include_column_name_suffixes=["_at", "age"])
        result = column_filter._apply_column_name_filters(sample_column_names)
        assert result == ["age", "created_at"]

    @pytest.mark.unit
    def test_apply_column_name_filters_exclude_suffixes(self, sample_column_names):
        """Test column name filtering with exclude suffixes."""
        column_filter = ColumnFilter(exclude_column_name_suffixes=["_at", "ion"])
        result = column_filter._apply_column_name_filters(sample_column_names)
        expected = ["id", "name", "age", "salary", "is_active"]
        assert result == expected

    @pytest.mark.unit
    def test_infer_semantic_type_numeric_integer(self, sample_column_types):
        """Test semantic type inference for integer columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(sample_column_types, "id")
        assert result == SemanticDomainTypes.NUMERIC

    @pytest.mark.unit
    def test_infer_semantic_type_numeric_decimal(self, sample_column_types):
        """Test semantic type inference for decimal columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(sample_column_types, "salary")
        assert result == SemanticDomainTypes.NUMERIC

    @pytest.mark.unit
    def test_infer_semantic_type_text(self, sample_column_types):
        """Test semantic type inference for text columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(sample_column_types, "name")
        assert result == SemanticDomainTypes.TEXT

    @pytest.mark.unit
    def test_infer_semantic_type_datetime(self, sample_column_types):
        """Test semantic type inference for datetime columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(
            sample_column_types, "created_at"
        )
        assert result == SemanticDomainTypes.DATETIME

    @pytest.mark.unit
    def test_infer_semantic_type_boolean(self, sample_column_types):
        """Test semantic type inference for boolean columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(
            sample_column_types, "is_active"
        )
        assert result == SemanticDomainTypes.LOGIC

    @pytest.mark.unit
    def test_infer_semantic_type_unknown(self, sample_column_types):
        """Test semantic type inference for unknown columns."""
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(
            sample_column_types, "nonexistent"
        )
        assert result == SemanticDomainTypes.UNKNOWN

    @pytest.mark.unit
    def test_infer_semantic_type_spark_backticks(self):
        """Test semantic type inference with Spark backtick column names."""
        column_types = [{"name": "`column_name`", "type": "INTEGER"}]
        column_filter = ColumnFilter()
        result = column_filter._infer_semantic_type_from_column_type(column_types, "column_name")
        assert result == SemanticDomainTypes.NUMERIC

    @pytest.mark.unit
    def test_build_semantic_type_map(self, mock_validator, sample_column_types):
        """Test building semantic type map."""
        mock_validator.get_metric.return_value = sample_column_types

        column_filter = ColumnFilter()
        column_names = ["id", "name", "created_at"]
        result = column_filter._build_semantic_type_map(mock_validator, column_names)

        expected = {
            "id": SemanticDomainTypes.NUMERIC,
            "name": SemanticDomainTypes.TEXT,
            "created_at": SemanticDomainTypes.DATETIME,
        }
        assert result == expected

    @pytest.mark.unit
    def test_apply_semantic_type_filters_include(self):
        """Test semantic type filtering with include types."""
        column_filter = ColumnFilter(include_semantic_types=[SemanticDomainTypes.NUMERIC])
        column_names = ["id", "name", "age"]
        semantic_type_map = {
            "id": SemanticDomainTypes.NUMERIC,
            "name": SemanticDomainTypes.TEXT,
            "age": SemanticDomainTypes.NUMERIC,
        }

        result = column_filter._apply_semantic_type_filters(column_names, semantic_type_map)
        assert result == ["id", "age"]

    @pytest.mark.unit
    def test_apply_semantic_type_filters_exclude(self):
        """Test semantic type filtering with exclude types."""
        column_filter = ColumnFilter(exclude_semantic_types=[SemanticDomainTypes.TEXT])
        column_names = ["id", "name", "age"]
        semantic_type_map = {
            "id": SemanticDomainTypes.NUMERIC,
            "name": SemanticDomainTypes.TEXT,
            "age": SemanticDomainTypes.NUMERIC,
        }

        result = column_filter._apply_semantic_type_filters(column_names, semantic_type_map)
        assert result == ["id", "age"]

    @pytest.mark.unit
    def test_get_filtered_column_names_no_semantic_filtering(
        self, mock_validator, sample_column_names
    ):
        """Test getting filtered column names without semantic filtering."""
        mock_validator.get_metric.return_value = sample_column_names

        column_filter = ColumnFilter(exclude_column_names=["description"])
        result = column_filter.get_filtered_column_names(mock_validator)

        expected = ["id", "name", "age", "salary", "created_at", "is_active"]
        assert result == expected

    @pytest.mark.unit
    def test_get_filtered_column_names_with_semantic_filtering(
        self, mock_validator, sample_column_names, sample_column_types
    ):
        """Test getting filtered column names with semantic filtering."""

        # Mock the validator to return column names and types
        def mock_get_metric(metric):
            if metric.metric_name == "table.columns":
                return sample_column_names
            elif metric.metric_name == "table.column_types":
                return sample_column_types
            return None

        mock_validator.get_metric.side_effect = mock_get_metric

        column_filter = ColumnFilter(
            include_semantic_types=[SemanticDomainTypes.NUMERIC], exclude_column_names=["id"]
        )
        result = column_filter.get_filtered_column_names(mock_validator)

        # Should include numeric columns (age, salary) but exclude 'id'
        expected = ["age", "salary"]
        assert result == expected

    @pytest.mark.unit
    def test_get_filtered_column_names_complex_filtering(
        self, mock_validator, sample_column_names, sample_column_types
    ):
        """Test getting filtered column names with complex filtering."""

        # Mock the validator to return column names and types
        def mock_get_metric(metric):
            if metric.metric_name == "table.columns":
                return sample_column_names
            elif metric.metric_name == "table.column_types":
                return sample_column_types
            return None

        mock_validator.get_metric.side_effect = mock_get_metric

        column_filter = ColumnFilter(
            include_semantic_types=[SemanticDomainTypes.NUMERIC, SemanticDomainTypes.TEXT],
            exclude_column_names=["description"],
            include_column_name_suffixes=["e", "y"],
        )
        result = column_filter.get_filtered_column_names(mock_validator)

        # Should include columns ending with 'e' or 'y' that are numeric or text,
        # but exclude 'description'
        # name (ends with 'e', TEXT), age (ends with 'e', NUMERIC), salary (ends with 'y', NUMERIC)
        expected = ["name", "age", "salary"]
        assert result == expected
